from time import time
from naeural_core.business.base.cv_plugin_executor import CVPluginExecutor

__VER__ = '2.1.0'

_CONFIG = {
  **CVPluginExecutor.CONFIG,
  
  'PROCESS_DELAY'           : 0.05,
  'ALLOW_EMPTY_INPUTS'      : True,
  'MAX_INPUTS_QUEUE_SIZE'   : 1, # default queue set to 1 due to imposed delay 
  
  'RESULT_KEY'              : 'exec_result',
  'ERROR_KEY'               : 'exec_errors',
  'WARNING_KEY'             : 'exec_warning',
    
  'CODE'        : "",
  'DEBUG_EXEC'  : False,
  
  'VALIDATION_RULES' : {
    **CVPluginExecutor.CONFIG['VALIDATION_RULES'],
    
    'CODE' : {
      'TYPE' : 'str',
    }
  },

}

SLEEP_ON_ERRORS = 10

class CustomExec01Plugin(CVPluginExecutor):
  CONFIG = _CONFIG
  def __init__(self, **kwargs):
    self.__first_run_executed = False
    self.__last_run = None
    self.__iteration = 0
    self.__null_payload_time = time()
    self.__debug_time = time()
    self.__debug_count = 0
    super().__init__(**kwargs)
    return
  
  
  @property
  def iteration(self):
    return self.__iteration
  
  def get_base64code(self):
    return self.cfg_code  

  def _process(self):
    self.__iteration += 1
    

    if not self.__first_run_executed:
      self.P("* * * * Custom executor up and running * * * * ", color='g')
      self.__first_run_executed = True      
    else:
      if self.cfg_debug_exec and (time() - self.__debug_time) > 10:
        self.P("DEBUG EXEC: * * * * Custom executor last exec (of {} execs) was ago {:.1f}s * * * *".format(
          self.__debug_count,
          time() - self.__last_run,
          ), color='y'
        )
        self.__debug_time = time()
        self.__debug_count = 0
      
    result, errors, payload, warnings = None, None, None, []
    
    self.__last_run = time()
    self.__debug_count += 1
    str_b64code = self.get_base64code()
    if len(str_b64code) > 0: # no need to test for `str` as is imposed by validation
      res = self.exec_code(
        str_b64code=str_b64code,
        debug=self.cfg_debug_exec,
        self_var='plugin', # we set plugin local for exec code
        modify=True,
      )
      if isinstance(res, tuple):
        result, errors, warnings = res
    
    if len(warnings) > 0:
      self.P("Execution warnings: {}".format(warnings), color='r')
      
    if errors is not None or result is not None:
      payload = self._create_payload(
        **{
          self.cfg_result_key  : result,
          self.cfg_error_key   : errors,
          self.cfg_warning_key : warnings,
          'exec_iter'          : self.__iteration,
        }
      )
      self.__null_payload_time = None
      if errors is not None:
        _show_err = True
        if isinstance(errors, list) and len(errors[-1]) > 100:
          trace = errors[-1]
          errors = errors[:-1]
        else:
          trace = None
        self.add_payload(payload)
        payload = None # reset as there is no need to send it again
        sep = '*' * 100
        msg = "Exception in {}".format(self)
        info = "Reset/sleep plugin for {}s. Error in custom code execution: {}\n{}\n{}\n{}\n{}".format(
          SLEEP_ON_ERRORS, errors, 
          sep, self.base64_to_code(str_b64code),
          sep if trace is not None else '', trace if trace is not None else ''
        )
        if _show_err:
          self.P(msg +' ' + info, color='r')
        self._create_error_notification(msg=msg, info=info, displayed=_show_err)
        self.reset_plugin_instance()
        self.sleep(SLEEP_ON_ERRORS)
      # endif we have errors
    # end if some data needs to be send
    else:
      if self.__null_payload_time is None:
        self.__null_payload_time = time()
      else:
        null_elapsed = time() - self.__null_payload_time
        if null_elapsed > self.cfg_send_manifest_each:
          self.__null_payload_time = time()
          msg = "Custom plugin {}:{}:{} was idle past {:.1f}s {}".format(
            self._stream_id,
            self._signature,
            self.cfg_instance_id,
            null_elapsed,
            '(IDLE DUE TO NO CODE PROVIDED)' if len(self.cfg_code) == 0 else '(code available)'
          )
          self.P(msg)
          payload = self._create_payload(
            **{
              self.cfg_result_key  : None,
              self.cfg_error_key   : msg,
              self.cfg_warning_key : None,
              'exec_iter'          : self.__iteration,
            }
          )
        #endif send idle message
      #endif check null payload time
    #endif something or nothing returned
    return payload
    


from kmonitor import KubeMonitor, safe_jsonify

if __name__ == '__main__':
  km = KubeMonitor()
  
  apps = ['emqx', 'nvidia', 'basic-test']
  
  result = km.check_pods_by_names(apps)  
  print("Pods {}:\n{}".format(apps, safe_jsonify(result, indent=2)))
  
  nodes = km.get_nodes_metrics()
  print("Nodes:\n{}".format(safe_jsonify(nodes, indent=2)))
  


# TODO: review if this is safe
from os import environ

from ratio1.const import (
  PAYLOAD_CT,
  COMMANDS,
  STATUS_TYPE,
  PAYLOAD_DATA,
  NET_CONFIG,
  CONFIG_STREAM,
  BIZ_PLUGIN_DATA,
  PLUGIN_INFO,
  COLORS,
  HB,
  BASE_CT,
  COMMS,
  NOTIFICATION_CODES,
  WEEKDAYS_SHORT,
)

EE_NETMON_ADDRESS_INDEX_ENV_KEY = "EE_NETMON_ADDRESS_INDEX"
EE_NETMON_SEND_ONLY_ONLINE_ENV_KEY = "EE_NETMON_SEND_ONLY_ONLINE"
EE_NETMON_SEND_CURRENT_NETWORK_EACH_ENV_KEY = "EE_NETMON_SEND_CURRENT_NETWORK_EACH"


EE_EPOCH_MANAGER_DEBUG = 'EE_EPOCH_MANAGER_DEBUG'
EE_GENESIS_EPOCH_DATE_KEY = 'EE_GENESIS_EPOCH_DATE'
EE_EPOCH_INTERVALS_KEY = 'EE_EPOCH_INTERVALS'
EE_EPOCH_INTERVAL_SECONDS_KEY = 'EE_EPOCH_INTERVAL_SECONDS'

###################################################
#############    Era information    ###############
###################################################

CURRENT_NETWORK = environ.get(BASE_CT.dAuth.DAUTH_NET_ENV_KEY, "devnet")
CURRENT_EVM_NET_DATA = BASE_CT.EVM_NET_DATA[CURRENT_NETWORK]
DEFAULT_GENESYS_EPOCH_DATE = CURRENT_EVM_NET_DATA[BASE_CT.EvmNetData.EE_GENESIS_EPOCH_DATE_KEY]
DEFAULT_EPOCH_INTERVALS = CURRENT_EVM_NET_DATA[BASE_CT.EvmNetData.EE_EPOCH_INTERVALS_KEY]
DEFAULT_EPOCH_INTERVAL_SECONDS = CURRENT_EVM_NET_DATA[BASE_CT.EvmNetData.EE_EPOCH_INTERVAL_SECONDS_KEY]
SUPERVISOR_MIN_AVAIL_PRC = CURRENT_EVM_NET_DATA[BASE_CT.EvmNetData.EE_SUPERVISOR_MIN_AVAIL_PRC_KEY]
EPOCH_MAX_VALUE = 255
###################################################
###################################################
###################################################

EE_ALIAS_MAX_SIZE = 15


ETH_ENABLED_ENV_KEY = BASE_CT.ETH_ENABLED_ENV_KEY # this env var tells if the system should use or not EVM compatibility in payloads
ETH_ENABLED = str(environ.get(ETH_ENABLED_ENV_KEY, 'False')).lower() == 'true'

LocalInfo = BASE_CT.LocalInfo

CONST_ADMIN_PIPELINE_NAME = 'admin_pipeline'

BLOCKCHAIN_MANAGER = 'BLOCKCHAIN_MANAGER'

R1FS_ENGINE = 'R1FS_ENGINE'

### IMPORTANT:
###  These are the plugin signatures that are not filtered by the admin-pipeline
###  payload filter. Only these payloads will be sent to the plugins
###
ADMIN_PIPELINE_FILTER = [
  "UPDATE_MONITOR_01", 
  "NET_MON_01", 
  "NET_CONFIG_MONITOR", 
  "CHAIN_STORE_BASE",
]

WHITELIST_COMMANDS_FILE = 'whitelist_commands.json'
TEMPLATE_WHITELIST_COMMANDS = [
  # !! NetMon queries
  # {
  #   "ACTION": COMMANDS.UPDATE_PIPELINE_INSTANCE,
  #   "PAYLOAD": {
  #     "SIGNATURE": "NET_MON_01",
  #     "INSTANCE_ID": "NET_MON_01_INST",
  #     "INSTANCE_CONFIG": {
  #       "INSTANCE_COMMAND": {
  #       }
  #     }
  #   }
  # },
  
  # # !! Get config startup
  # {
  #   "ACTION": COMMANDS.UPDATE_PIPELINE_INSTANCE,
  #   "PAYLOAD": {
  #     "SIGNATURE": "UPDATE_MONITOR_01",
  #     "INSTANCE_CONFIG": {
  #       "INSTANCE_COMMAND": {
  #         "COMMAND": "GET_CONFIG",
  #       }
  #     }
  #   },
  # },
  # !! Edit config startup - requires authorization
  # {
  #   "ACTION": COMMANDS.UPDATE_PIPELINE_INSTANCE,
  #   "PAYLOAD": {
  #     "SIGNATURE": "UPDATE_MONITOR_01",
  #     "INSTANCE_CONFIG": {
  #       "INSTANCE_COMMAND": {
  #         "COMMAND": "SAVE_CONFIG",
  #       }
  #     }
  #   },
  # },

  # !! Restart Edge Node - requires authorization
  # {
  #   "ACTION": COMMANDS.STOP,
  # },

  # !! Restart OS - requires authorization
  # {
  #   "ACTION": COMMANDS.RESTART,
  # },

  # Request heartbeat with extra info
  {
    "ACTION": COMMANDS.FULL_HEARTBEAT,
  },
  {
    "ACTION": COMMANDS.TIMERS_ONLY_HEARTBEAT,
  },
]

ADMIN_PIPELINE = {
  "PLUGIN_MONITOR_01":{
    "PROCESS_DELAY": 1,
  },

  "LOCAL_COMMS_01": {
    "LOCAL_COMMS_ENABLED_ON_STARTUP": False
  },

  
  # "MINIO_MONIT_01": {
  #   "PROCESS_DELAY": 20,
  #   "MINIO_HOST": None,
  #   "MINIO_ACCESS_KEY": None,
  #   "MINIO_SECRET_KEY": None,
  #   "MINIO_SECURE": None,
  #   "MAX_SERVER_QUOTA"  : 95
  # },

  # "REST_CUSTOM_EXEC_01": {
  #   "ALLOW_EMPTY_INPUTS": True,
  #   "RUN_WITHOUT_IMAGE": True,
  #   "SEND_MANIFEST_EACH": 301
  # },

  # "SELF_CHECK_01": {
  #   "DISK_LOW_PRC": 0.15,
  #   "MEM_LOW_PRC": 0.15,
  #   "PROCESS_DELAY": 20
  # },
  
  "NET_CONFIG_MONITOR" : {
    "PROCESS_DELAY": 0,
    "VERBOSE_NETCONFIG_LOGS" : False,
  },
  
  "CHAIN_STORE_BASE" : {
    "PROCESS_DELAY": 0,
    "FULL_DEBUG_PAYLOADS" : False,
    "CHAIN_STORE_DEBUG" : False, 
  },

  "NET_MON_01": {
    "PROCESS_DELAY": 20,
    "LOG_INFO"            : False,
    "LOG_FULL_INFO"       : False,
    "SUPERVISOR": "$EE_SUPERVISOR" # obsolete/redundant as of 2025-03-03
  },
  

  "UPDATE_MONITOR_01": {
    "PROCESS_DELAY": 300,
    # "WORKING_HOURS": [],
    "VERSION_TOKEN": None,
    "RESTART_ON_BEHIND": True,
    "VERSION_URL": "https://raw.githubusercontent.com/Ratio1/edge_node/ver.py"
  },
  
  # "K8S_MONITOR_01": {
  #   "PROCESS_DELAY": 10,
  # },

  # "SYSTEM_HEALTH_MONITOR_01": {
  #   "PROCESS_DELAY": 180,
  #   "KERNEL_LOG_LEVEL" : "emerg,alert,crit,err"
  # }
}

ADMIN_PIPELINE_NETMON = 'NET_MON_01'

NOTIFICATION_TYPE = STATUS_TYPE

NOTIFICATION = PAYLOAD_DATA.NOTIFICATION

EE_ID = BASE_CT.EE_ID
EE_ADDR = HB.EE_ADDR

NON_ACTIONABLE_INSTANCE_CONFIG_KEYS = [
  'INSTANCE_COMMAND_LAST',
]


LOCAL_CACHE = '_local_cache'

CONFIG_APP_FALLBACK = 'CONFIG_APP_FALLBACK'
CONFIG_APP_DOCKER_FALLBACK = './scripts/docker_env_ext/config_app.txt'

# CV payload
TLBR_POS = PAYLOAD_CT.TLBR_POS
PROB_PRC = PAYLOAD_CT.PROB_PRC
TYPE = PAYLOAD_CT.TYPE
COLOR_TAG = PAYLOAD_CT.COLOR_TAG


PLUGIN_END_PREFIX = ">>>> STOP: "
BM_PLUGIN_END_PREFIX = ">> STOP: "


class CALLBACKS:
  INSTANCE_CONFIG_SAVER_CALLBACK = 'instance_config_saver_callback'
  PIPELINE_CONFIG_SAVER_CALLBACK = 'pipeline_config_saver_callback'
  MAIN_LOOP_RESOLUTION_CALLBACK = 'main_loop_resolution_callback'


class IPFS:
  UPLOAD_CONFIG = 'UPLOAD_CONFIG'


class SERVING:
  SERVER_COLLECTOR_TIMEDELTA = 'SERVER_COLLECTOR_TIMEDELTA'
  SERVER_COLLECTOR_DELAY = 'SERVER_COLLECTOR_DELAY'
  CHECK_BLOCKED_INPROCESS_SERVING = 'CHECK_BLOCKED_INPROCESS_SERVING'
  MAX_WAIT_TIME_MULTIPLIER = 'MAX_WAIT_TIME_MULTIPLIER'
  SHM_IMG_MAX_SHAPE = (1520, 2688, 3)
  SHM_MAX_LEN = 50
  COMM_METHOD = 'COMM_METHOD'


class CONFIG_APP_v2:
  K_COMMUNICATION = 'COMMUNICATION'


class CONFIG_MANAGER:
  DEFAULT_SUBFOLDER_PATH = 'box_configuration'
  DEFAULT_FN_APP_CONFIG = 'config_app.txt'
  DEFAULT_FOLDER_STREAMS_CONFIGS = 'streams'


class CONFIG_INSTANCE:
  K_INSTANCE_ID = BIZ_PLUGIN_DATA.INSTANCE_ID


class CONFIG_INSTANCE_COMPUTER_VISION(CONFIG_INSTANCE):
  K_COORDS = BIZ_PLUGIN_DATA.COORDS
  K_POINTS = BIZ_PLUGIN_DATA.POINTS


class CONFIG_PLUGIN:
  K_INSTANCES = BIZ_PLUGIN_DATA.INSTANCES
  K_SIGNATURE = BIZ_PLUGIN_DATA.SIGNATURE


class CONFIG_RETRIEVE:
  K_TYPE = 'TYPE'
  K_CONNECT_PARAMS = 'CONNECT_PARAMS'
  K_APP_CONFIG_ENDPOINT = 'APP_CONFIG_ENDPOINT'
  K_STREAMS_CONFIGS_ENDPOINT = 'STREAMS_CONFIGS_ENDPOINT'


class CONFIG_STARTUP_v2:
  K_EE_ID = 'EE_ID'
  K_LITE = 'LITE'
  K_CONFIG_RETRIEVE = 'CONFIG_RETRIEVE'
  DEFAULT_EE_ID = 'XXXXXXXXXX'
  K_SECURED = 'SECURED'
  SECURED = K_SECURED
  BLOCKCHAIN_CONFIG = 'BLOCKCHAIN_CONFIG'
  CHECK_RAM_ON_SHUTDOWN = 'CHECK_RAM_ON_SHUTDOWN'


class PLUGIN_SEARCH:
  # SEARCH_PACKAGES
  SEARCH_IN_PACKAGES = ["ratio1", "naeural_core"]

  # CONFIG
  LOC_CONFIG_RETRIEVE_PLUGINS = 'plugins.config'
  SUFFIX_CONFIG_RETRIEVE_PLUGINS = 'ConfigRetriever'
  SAFE_LOC_CONFIG_PLUGINS = ['naeural_core.config.default']

  # COMM
  LOC_COMM_PLUGINS = 'plugins.comm'
  SUFFIX_COMM_PLUGINS = 'CommThread'
  SAFE_LOC_COMM_PLUGINS = ['naeural_core.comm.default',]

  # DATA
  LOC_DATA_ACQUISITION_PLUGINS = ['plugins.data', 'extensions.data']
  SAFE_LOC_DATA_ACQUISITION_PLUGINS = ['naeural_core.data.default', 'naeural_core.data.training', 'extensions.data']
  SUFFIX_DATA_ACQUISITION_PLUGINS = 'DataCapture'
  SAFE_LOC_DATA_ACQUISITION_IMPORTS = ['naeural_core.data', 'extensions.data']

  # SERVING
  LOC_SERVING_PLUGINS = ['plugins.serving.inference', 'plugins.serving.training']
  SUFFIX_SERVING_PLUGINS = None
  SAFE_LOC_SERVING_PLUGINS = ['naeural_core.serving.default_inference', 'naeural_core.serving.training', 'extensions.serving']
  SERVING_SAFE_IMPORTS = ['naeural_core.serving', 'extensions.serving']  # be aware that someone may drop some .py there and "validate" attacker code

  # BUSINESS
  LOC_BIZ_PLUGINS = ['plugins.business']
  SAFE_BIZ_PLUGINS = ['naeural_core.business.default', 'naeural_core.business.training', 'extensions.business']  # this is were the safe plugins are stored
  SAFE_BIZ_IMPORTS = ['naeural_core.business', 'extensions.business']  # be aware that someone may drop some .py there and "validate" attacker code
  SUFFIX_BIZ_PLUGINS = 'Plugin'

  # Heavy Ops
  SAFE_LOC_HEAVY_OPS_PLUGINS = 'naeural_core.heavy_ops.default'
  LOC_HEAVY_OPS_PLUGINS = 'plugins.heavy_ops'
  SUFFIX_HEAVY_OPS_PLUGINS = 'HeavyOp'

  # FILE SYS
  LOC_FILE_SYSTEM_PLUGINS = 'plugins.remote_file_system'
  SUFFIX_FILE_SYSTEM_PLUGINS = 'FileSystem'
  LOC_SAFE_FILE_SYSTEM_PLUGINS = ['naeural_core.remote_file_system.default']

  # TESTING FRAMEWORK
  LOC_TESTING_FRAMEWORK_TESTING_PLUGINS = 'plugins.business.test_framework.testing'
  SUFFIX_TESTING_FRAMEWORK_TESTING_PLUGINS = 'TestingPlugin'

  LOC_TESTING_FRAMEWORK_SCORING_PLUGINS = 'plugins.business.test_framework.scoring'
  SUFFIX_TESTING_FRAMEWORK_SCORING_PLUGINS = 'ScoringPlugin'


class RETURN_CODES:
  CODE_RESTART = 10
  CODE_EXCEPTION = 11
  CODE_STOP = 12


class SYS_MON:
  NAME = 'SYSTEM_MONITOR'


class EMAIL_NOTIFICATION:
  EMAIL_CONFIG = 'EMAIL_CONFIG'
  DEFAULT_EMAIL_CONFIG = None
  SERVER = 'SERVER'
  DEFAULT_SERVER = ''
  PORT = 'PORT'
  DEFAULT_PORT = 25
  USER = 'USER'
  DEFAULT_USER = 'tmp@email.com'
  PASSWORD = 'PASSWORD'
  DEFAULT_PASSWORD = ''
  DESTINATION = 'DESTINATION'


class IMAGE_COMPRESSION:
  MAX_HEIGHT = 720
  QUALITY = 60
  ORIG_MAX_HEIGHT = 1080


class MINIO:
  ACCESS_KEY = 'access_key'
  SECRET_KEY = 'secret_key'
  BUCKET_NAME = 'bucket_name'
  ENDPOINT = 'endpoint'
  FILE_RETENTION = 'file_retention'
  SECURE = 'secure'


class SHMEM:
  BUSINESS = 'BUSINESS'
  SERVING = 'SERVING'
  DCT = 'DCT'
  COMM = 'COMM'


class GPU_INFO:
  INFO = 'GPU_INFO'
  NAME = 'NAME'
  TOTAL_MEM = 'TOTAL_MEM'
  GPU_USED = 'GPU_USED'
  ALLOCATED_MEM = 'ALLOCATED_MEM'
  GPU_TEMP = "GPU_TEMP"
  GPU_FAN_SPEED = "GPU_FAN_SPEED"
  GPU_FAN_SPEED_UNIT = "GPU_FAN_SPEED_UNIT"


class MOVIE_EVAL_STAGES:
  INFERENCE = 'INFERENCE'
  WRITE = 'WRITE'


class FILE_SIZE_UNIT:
  BYTES = 'BYTES'
  KB = 'KB'
  MB = 'MB'
  GB = 'GB'


CUSTOM_DOWNLOADABLE_MODEL = 'CUSTOM_DOWNLOADABLE_MODEL'
CUSTOM_DOWNLOADABLE_MODEL_URL = 'CUSTOM_DOWNLOADABLE_MODEL_URL'
MODEL_ZOO_CONFIG = 'MODEL_ZOO_CONFIG'
SERVING_TIMERS_IDLE_DUMP = 'SERVING_TIMERS_IDLE_DUMP'
SERVING_TIMERS_PREDICT_DUMP = 'SERVING_TIMERS_PREDICT_DUMP'

# root engine version
LITE = 'LITE'


# communication
COMMUNICATORS = 'COMMUNICATORS'
COMMUNICATOR_REST = 'REST'
COMMUNICATOR_PIKA = 'PIKA'
COMMUNICATOR_PAHO = 'PAHO'

CONN_MAX_RETRY_ITERS = 'CONN_MAX_RETRY_ITERS'

AVAILABLE_STREAMS_PICTURES = 'AVAILABLE_STREAMS_PICTURES'
CHECK = 'CHECK'
EXCEPTION = 'EXCEPTION'
HEARTBEAT = 'HEARTBEAT'
PAYLOAD = 'PAYLOAD'
RESTART = 'RESTART'
STATUS = 'STATUS'
STOP = 'STOP'
UPDATE_CONFIG = 'UPDATE_CONFIG'
DEVICE_STATUS = 'DEVICE_STATUS'
DEVICE_MESSAGE = 'DEVICE_MESSAGE'

DEVICE_LOG = 'DEVICE_LOG'
ERROR_LOG = 'ERROR_LOG'

# DEVICE STATUS
DEVICE_STATUS_AVAILABLE_STREAMS_PICTURES = 'AVAILABLE_STREAMS_PICTURES'
DEVICE_STATUS_CHECK = 'CHECK'
DEVICE_STATUS_EXCEPTION = 'EXCEPTION'
DEVICE_STATUS_ONLINE = 'ONLINE'
DEVICE_STATUS_RESTART = 'RESTART'
DEVICE_STATUS_SHUTDOWN = 'SHUTDOWN'
DEVICE_STATUS_STOP = 'STOP'
DEVICE_STATUS_UPDATE_CONFIG = 'UPDATE_CONFIG'

ID = 'ID'
ACTION = 'ACTION'

MQ_BROKER = 'MQ_BROKER'
MQ_PORT = 'MQ_PORT'
MQ_USER = 'MQ_USER'
MQ_PASS = 'MQ_PASS'
MQ_PATH = 'MQ_PATH'

MQ_CONFIG_EXCHANGE = 'CONFIG_EXCHANGE'
MQ_CONFIG_EXCHANGE_TYPE = 'CONFIG_EXCHANGE_TYPE'
MQ_CONFIG_QUEUE = 'CONFIG_QUEUE'
MQ_CONFIG_ROUTING_KEY = 'CONFIG_ROUTING_KEY'
MQ_CONFIG_QUEUE_DURABLE = 'CONFIG_QUEUE_DURABLE'
MQ_CONFIG_QUEUE_EXCLUSIVE = 'CONFIG_QUEUE_EXCLUSIVE'
MQ_CONFIG_QUEUE_DEVICE_SPECIFIC = 'CONFIG_QUEUE_DEVICE_SPECIFIC'


DISABLE_THREAD = 'DISABLE_THREAD'


MQ_PAYLOADS_EXCHANGE = 'PAYLOADS_EXCHANGE'
MQ_PAYLOADS_EXCHANGE_TYPE = 'PAYLOADS_EXCHANGE_TYPE'
MQ_PAYLOADS_QUEUE = 'PAYLOADS_QUEUE'
MQ_PAYLOADS_ROUTING_KEY = 'PAYLOADS_ROUTING_KEY'
MQ_PAYLOADS_QUEUE_DURABLE = 'PAYLOADS_QUEUE_DURABLE'
MQ_PAYLOADS_QUEUE_EXCLUSIVE = 'PAYLOADS_QUEUE_EXCLUSIVE'
MQ_PAYLOADS_QUEUE_DEVICE_SPECIFIC = 'PAYLOADS_QUEUE_DEVICE_SPECIFIC'

COMMUNICATION = 'COMMUNICATION'
NR_PAYLOADS_TO_SEND = 'NR_PAYLOADS_TO_SEND'
BUFFER_MAX_LEN = 'BUFFER_MAX_LEN'


# config KEYS
CONFIG_APP = 'CONFIG_APP'
CONFIG_PLUGINS = 'CONFIG_PLUGINS'
CONFIG_INFERENCE = 'CONFIG_INFERENCE'
CONFIG_INFERENCE_BENCHMARK = 'CONFIG_INFERENCE_BENCHMARK'
CONFIG_STREAMS = 'CONFIG_STREAMS'
CONFIG_DEBUG_WINDOW = 'CONFIG_DEBUG_WINDOW'
CONFIG_STARTUP = 'CONFIG_STARTUP'
CONFIG_MODELS_ZOO = 'CONFIG_MODELS_ZOO'
CONFIG_MODELS_ZOO_BENCHMARK = 'CONFIG_MODELS_ZOO_BENCHMARK'


# RETURN CODES
CODE_STOP = 0
CORE_NODE_RESTART = CODE_STOP
CODE_SHUTDOWN = 1
CODE_RESTART = 10
CODE_BOX_RESTART = CODE_RESTART
CODE_EXCEPTION = 51
CODE_CONFIG_ERROR = 52


# STREAMS STATES
STATE_ALIVE = 'ALIVE'
STATE_EXCEPTION = 'EXCEPTION'
STATE_FINISHED = 'FINISHED'
STATE_STOPPED = 'STOPPED'
STATE_KILLED = 'KILLED'

# STATUS TYPES
STATUS_NORMAL = 'NORMAL'
STATUS_EXCEPTION = 'EXCEPTION'

# TIMERS
TIMER_APP = 'app_alive_timer'
TIMER_ASCONTIGUOUSARRAY = 'ascontiguousarray'
TIMER_BGR_WITNESS = 'bgr_witness'
TIMER_CA_ANCHOR = 'ca_anchor'
TIMER_CA_ANCHOR_MEAN = 'ca_anchor_mean'
TIMER_CA_COLLECT_ANCHOR = 'ca_collect_anchor'
TIMER_CA_COLLECT_CURRENT = 'ca_collect_current'
TIMER_CA_CREATE_ANCHOR_MEAN = 'ca_create_anchor_mean'
TIMER_CA_CREATE_ANCHOR = 'ca_create_anchor'
TIMER_CA_CREATE_CURRENT = 'ca_create_current'
TIMER_CA_CREATE_CURRENT_MEAN = 'ca_create_current_mean'
TIMER_CA_EVALUATE = 'ca_evaluate'
TIMER_CA_SSIM = 'ca_ssim'
TIMER_CA_SSIM_OPERATION = 'ca_ssim_operation'
TIMER_CA_GET_SURF_FEATURES = 'ca_surf'
TIMER_CA_MATCH_KEYPOINTS = 'ca_match_kpts'
TIMER_CA_HOMOGRAPHY = 'ca_homography'
TIMER_CA_GET_DISTANCE_PLUS_LABELS = 'ca_eval_dist_lbl'
TIMER_CA_GET_DISTANCE = 'ca_eval_dist'
TIMER_COPY_WITNESS = 'copy_witness'
TIMER_CQC = 'cqc'
TIMER_CQC_COLLECT = 'cqc_collect'
TIMER_CQC_EVALUATE = 'cqc_evaluate'
TIMER_GET_CAPTURES_DATA = 'get_captures_data'
TIMER_GET_CAPTURES_MESSAGES = 'get_captures_messages'
TIMER_GET_WITNESS_IMAGE = 'get_witness_image'
TIMER_INFERENCE = 'inference'
TIMER_INFER = 'infer'
TIMER_MAIN_LOOP = 'main_loop'
TIMER_PREPARE_PAYLOAD = 'prepare_payload'
TIMER_PREPARE_RESULTS = 'prepare_results'
TIMER_PRE_PROCESS = 'TIMER_PRE_PROCESS'
TIMER_PREPROCESS_WITNESS = 'pre_process_witness'
TIMER_POSTPROCESS_WITNESS = 'post_process_witness'
TIMER_REFRESH_APP = 'refresh_app'
TIMER_REFRESH_PLUGINS = 'refresh_plugins'
TIMER_REFRESH_STREAMS = 'refresh_streams'
TIMER_RUN_PLUGINS = 'run_plugins'
TIMER_SAVE_BUFFER_TO_DISK = 'save_buffer_to_disk'
TIMER_SEND_BUFFER_PAYLOAD = 'send_buffer_payload'
TIMER_SEND_INSTANT_PAYLOAD = 'send_instant_payload'
TIMER_SEND_PRIORITY_PAYLOAD = 'send_priority_payload'
TIMER_SET_STREAM_CONFIG = 'set_stream_config'
TIMER_WITNESS_BASE64 = 'witness_base64'


# COMMON STRINGS
ACTION_ADD = 'ADD'
ACTION_DELETE = 'DELETE'
ACTION_UPDATE = 'UPDATE'
ACCESS_TOKEN = 'ACCESS_TOKEN'
ACCESS_ALLOWED = 'ACCESS_ALLOWED'
ACCESS_ALLOWED_TIME = 'ACCESS_ALLOWED_TIME'
ACCESS_ALLOWED_NOTIFIED = 'ACCESS_ALLOWED_NOTIFIED'
ACTIVE_PLUGINS = 'ACTIVE_PLUGINS'
AI_ENGINE = 'AI_ENGINE'
ALERT_TYPE = 'ALERT_TYPE'
ALERT_THRESHOLD = 'ALERT_THRESHOLD'
ALERT_MODE = 'ALERT_MODE'
ALERT_RAISE_VALUE = 'ALERT_RAISE_VALUE'
ALERT_LOWER_VALUE = 'ALERT_LOWER_VALUE'
ALERT_HELPER = 'ALERT_HELPER'
ALIVE_TIME_MINS = 'ALIVE_TIME_MINS'
ALLOCATED_MEM = 'ALLOCATED_MEM'
ALERT_OBJECTS = 'ALERT_OBJECTS'


ANOMALY_PROBA = 'ANOMALY_PROBA'
ANOMALY_THRESHOLD = 'ANOMALY_THRESHOLD'
ANGLE = 'ANGLE'
ANGLE_LOW = 'ANGLE_LOW'
ANGLE_HIGH = 'ANGLE_HIGH'
ANGLE_ANCHORS = 'ANGLE_ANCHORS'
ANGLE_RESIZE = 'ANGLE_RESIZE'
ANGLE_SSIM_INTERVALS = 'ANGLE_SSIM_INTERVALS'
ANGLE_THRESHOLD = 'ANGLE_THRESHOLD'
APPEARANCES = 'APPEARANCES'
APPLICATION = 'APPLICATION'
APPLICATION_STATUS = 'APPLICATION_STATUS'
APPMON_VERSION = 'APPMON_VERSION'
ARRIVAL_TIME = 'ARRIVAL_TIME'
AVAILABLE_MEMORY = 'AVAILABLE_MEMORY'
AVAILABLE_DISK = 'AVAILABLE_DISK'
AVERAGE = 'AVERAGE'
AUDIT_PLUGINS = 'audit_plugins'
BASIC_OBSTRUCTION_MIN_CONFIDENCE = 'BASIC_OBSTRUCTION_MIN_CONFIDENCE'
BATCH_STRATEGY = 'BATCH_STRATEGY'
BATCH_STRATEGY_MOST_COMMON_SHAPE = 'BATCH_STRATEGY_MOST_COMMON_SHAPE'
BG_SUBTRACTOR_VAR_THRS = 'BG_SUBTRACTOR_VAR_THRS'
BIG_CIRCLE_RADIUS_PERCENT = 'BIG_CIRCLE_RADIUS_PERCENT'
BLUR = 'BLUR'
BLUR_ADAPTIVE = 'BLUR_ADAPTIVE'
BLUR_ON = 'BLUR_ON'
BLUR_RAW = 'BLUR_RAW'
BLUR_ZONES = 'BLUR_ZONES'
BOTH = 'BOTH'
BOTTOM = 'BOTTOM'
BOX_OFFSET = 'BOX_OFFSET'
CA_ALERT = 'CA_ALERT'
CA_ANCHOR = 'CA_ANCHOR'
CA_ANCHORS = 'CA_ANCHORS'
CA_ANCHOR_SOURCE = 'CA_ANCHOR_SOURCE'
CA_CURRENT = 'CA_CURRENT'
CA_CURRENT_SOURCE = 'CA_CURRENT_SOURCE'
CA_IMAGE = 'CA_IMAGE'
CA_STATE = 'CA_STATE'
CA_SSIM = 'CA_SSIM'
CA_SSIM_LABEL = 'CA_SSIM_LABEL'
CA_DIST = 'CA_DIST'
CA_DIST_LABEL = 'CA_DIST_LABEL'
CACHE = 'CACHE'

CACHE_DATA_CAPTURE = 'cache_data_capture'
CACHE_PLUGINS = 'cache_plugins'
CACHE_SERVING = 'cache_serving'

CLASSES = 'CLASSES'
CLOSE = 'CLOSE'
COLOR = 'COLOR'
COLOR_RESOLUTION_INTERVAL = 'COLOR_RESOLUTION_INTERVAL'
COLLECTOR = 'COLLECTOR'
COMM_TYPE = 'COMM_TYPE'
CONFIG = 'CONFIG'
CONFIDENCE_THRESHOLD = 'CONFIDENCE_THRESHOLD'
COORDONATES = 'COORDONATES'
COUNT = 'COUNT'
CPU = 'CPU'
COVERED_SERVERS = 'COVERED_SERVERS'
CPU_USED = 'CPU_USED'
CQC_ALERT = 'CQC_ALERT'  # check if in use
CQC_CLASSES = 'CQC_CLASSES'  # check if in use
CQC_DATA = 'CQC_DATA'  # check if in use
CQC_INFERENCE = 'CQC_INFERENCE'  # check if in use
CQC_LABEL = 'CQC_LABEL'  # check if in use
CQC_PROC = 'CQC_PROC'  # check if in use
CQC_PROBS = 'CQC_PROBS'  # check if in use
CQC_OBSTRUCTION = 'OBSTRUCTION'


CURRENT_TIME = 'CURRENT_TIME'
COMMUNICATION_THREAD = 'Thread'
COMMAND = 'COMMAND'
CONFIG_RETRIEVER = 'ConfigRetriever'
COORDS = 'COORDS'
COORDS_NONE = 'NONE'
COORDS_POINTS = 'POINTS'
COORDS_TLBR = 'TLBR'
LOOPS_TIMINGS = 'LOOPS_TIMINGS'
DAILY_INTERVALS = 'DAILY_INTERVALS'
DATA = 'DATA'
DATA_ACQUISITION = 'DATA_ACQUISITION'
DATA_HELPER_NAME = 'DATA_HELPER_NAME'
DATA_HELPER_PARAMS = 'DATA_HELPER_PARAMS'
DATA_CAPTURE_THREAD = 'DataCaptureThread'
DATA_CAPTURE = 'DataCapture'
DATE = 'DATE'
DEFAULT_EE_ID = 'XXXXXXXXXX'
DISABLE = 'DISABLE'
DEBUG = 'DEBUG'
DEBUG_DATA = 'DEBUG_DATA'
DEBUG_INFO = 'DEBUG_INFO'
DEBUG_PROP = 'DEBUG_PROP'
DEBUG_META = 'DEBUG_META'
DEBUG_TIMERS = 'DEBUG_TIMERS'
DEBUG_WINDOW = 'DEBUG_WINDOW'
DEBUG_OBJECTS = 'DEBUG_OBJECTS'
DEBUG_OBJECTS_SUMMARY = 'DEBUG_OBJECTS_SUMMARY'
DEBUG_PAYLOADS = 'DEBUG_PAYLOADS'
DEFAULT_CUDA = 'DEFAULT_DEVICE'
DEMO_MODE = 'DEMO_MODE'
DESCRIPTION = 'DESCRIPTION'
DETECTIONS = 'DETECTIONS'
DEVICE_ID = 'DEVICE_ID'
DOWNLOAD = 'DOWNLOAD'
DOWNLOAD_ELAPSED_TIME = 'DOWNLOAD_ELAPSED_TIME'
DOWNLOAD_PROGRESS = 'DOWNLOAD_PROGRESS'
DOWNLOAD_START_DATE = 'DOWNLOAD_START_DATE'
DOWNLOADS = 'DOWNLOADS'
DRAW_CACHE = 'DRAW_CACHE'
DROPBOX = 'DROPBOX'
ELAPSED_TIME = 'ELAPSED_TIME'
ENABLE = 'ENABLE'
ENABLE_QUALITY_CHECK = 'ENABLE_QUALITY_CHECK'
ENABLE_CAMERA_ANGLE = 'ENABLE_CAMERA_ANGLE'
ENABLE_CAMERA_OBSTRUCTION = 'ENABLE_CAMERA_OBSTRUCTION'
ENABLE_HEURISTIC_CAMERA_OBSTRUCTION = 'ENABLE_HEURISTIC_CAMERA_OBSTRUCTION'
ENABLE_MODEL_BASED_CAMERA_OBSTRUCTION = 'ENABLE_MODEL_BASED_CAMERA_OBSTRUCTION'
ENABLED = 'ENABLED'
END = 'END'
ENTROPY_DISK_RADIUS = 'ENTROPY_DISK_RADIUS'
ENTROPY_RATIO_BREAKPOINT = 'ENTROPY_RATIO_BREAKPOINT'
ERROR = 'ERROR'
EXCLUDE_ZONES = 'EXCLUDE_ZONES'
EXP_DT = 'EXP_DT'
EXP_ST = 'EXP_ST'
EXPIRATION_SECONDS = 'EXPIRATION_SECONDS'
EXCEEDING_STANDING_STILL = 'EXCEEDING_STANDING_STILL'
EXCEEDING_STANDING_STILL_TIME = 'EXCEEDING_STANDING_STILL_TIME'
EXCEEDING_STANDING_STILL_NOTIFIED = 'EXCEEDING_STANDING_STILL_NOTIFIED'
EXCLUDE_OBJECTS = 'EXCLUDE_OBJECTS'
FA_FNA = 'FA_FNA'
FILE_UPLOAD = 'FILE_UPLOAD'
FIXED_ZONES = 'FIXED_ZONES'
FOLDER_DATA = 'data'
FOLDER_MODELS = 'models'
FOLDER_NAME = 'FOLDER_NAME'
FOLDER_OUTPUT = 'output'
FOLDER_LOGS = 'logs'
FOLDER_DOWNLOADS = 'downloads'
FORCE_UPDATE = 'FORCE_UPDATE'
FPS = 'FPS'
FPS_OUTPUT = 'FPS_OUTPUT'
FRAME_H = 'FRAME_H'
FRAME_W = 'FRAME_W'
FRAME_CURRENT = 'FRAME_CURRENT'
FRAME_COUNT = 'FRAME_COUNT'
FRAME_INTERVAL = 'FRAME_INTERVAL'
FRAMES_INTERVAL = 'FRAMES_INTERVAL'
FRAMEWORK = 'FRAMEWORK'
FULL = 'FULL'
FULL_BLUR = 'FULL_BLUR'
GIT_BRANCH = 'GIT_BRANCH'
GPUS = 'GPUS'
GPU_MIN_MEMORY_MB = 'GPU_MIN_MEMORY_MB'
GOOD = 'GOOD'
GRAPH = 'GRAPH'
GRAPH_TYPE = 'GRAPH_TYPE'
GQL = 'GQL'
HAS_FINISHED = 'HAS_FINISHED'
HAS_PAYLOAD = 'HAS_PAYLOAD'
HEATMAP_RECORD_INTERVAL = 'HEATMAP_RECORD_INTERVAL'
HISTORY = 'HISTORY'
HOST = 'HOST'

IMG = 'IMG'
IMAGE = 'IMAGE'
IMAGE_HW = 'IMAGE_HW'
INSTANCES = 'INSTANCES'
IN_CONFIRMATION = 'IN_CONFIRMATION'
INCREASE_WITH_PIXELS = 'INCREASE_WITH_PIXELS'
INFER = 'INFER'
INFERENCE = 'INFERENCE'
INSTANCE_INFERENCES = 'INSTANCE_INFERENCES'
INFERENCE_EXECUTOR = 'INFERENCE_EXECUTOR'
INFERENCE_GRAPHS = 'INFERENCE_GRAPHS'
INFERENCE_GRAPHS_OUTPUT = 'INFERENCE_GRAPHS_OUTPUT'
INFERENCE_META_GRAPHS = 'INFERENCE_META_GRAPHS'
INFERENCE_ON_FULL_IMAGE = 'INFERENCE_ON_FULL_IMAGE'
INFERENCE_TS = 'INFERENCE_TS'
INFERENCE_GRAPHS = 'INFERENCE_GRAPHS'
INFER_INTERSECT = 'INFER_INTERSECT'
INSTANCES = 'INSTANCES'
INSTANCE = 'INSTANCE'
INSTANCE_ID = 'INSTANCE_ID'
INTERVAL_STATE = 'INTERVAL_STATE'
INTERSECTS = 'INTERSECTS'
INSTANCE_ID = 'INSTANCE_ID'
IO_FORMATTER = 'IO_FORMATTER'
IREGULAR = 'IREGULAR'
IS_ALERT = 'IS_ALERT'
IS_THREAD = 'IS_THREAD'
IS_NEW_ALERT = 'IS_NEW_ALERT'
ITERATION = 'ITERATION'
IS_ANOM_PRESENCE = 'IS_ANOM_PRESENCE'
IS_ANOM_ENVIRON = 'IS_ANOM_ENVIRON'
TAGS = 'TAGS'

JETSON_VIDEO_STREAM = 'JETSON_VIDEO_STREAM'
LAST_COUNT = 'LAST_COUNT'
LAST_EXPORTED_STATE = 'LAST_EXPORTED_STATE'
LAST_EXPORTED_DATE = 'LAST_EXPORTED_DATE'
LAST_EXPORT = 'LAST_EXPORT'
LAST_READ = 'LAST_READ'
LAST_SEEN_TIME = 'LAST_SEEN_TIME'
LAST_TEMPERATURE = 'LAST_TEMPERATURE'
LAST_HUMIDITY = 'LAST_HUMIDITY'
LAST_STRUCT_TIME = 'LAST_STRUCT_TIME'
LAST_PAYLOAD = 'LAST_PAYLOAD'
LINE = 'LINE'
LEFT = 'LEFT'
LICENSE_PLATE_TRACKER = 'LICENSE_PLATE_TRACKER'
LIVE_FEED = 'LIVE_FEED'
LOCATION = 'LOCATION'
LOCATION_NAME = 'LOCATION_NAME'
LOCATIONS = 'LOCATIONS'
LOCK_CMD = 'lock_cmd'
MAIN = 'MAIN'
MASTER = 'MASTER'
MACHINE_IP = 'MACHINE_IP'
MACHINE_MEMORY = 'MACHINE_MEMORY'
MAX_IDLE_TIME = 'MAX_IDLE_TIME'
MESSAGE = 'MESSAGE'
MESSAGE_ID = 'MESSAGE_ID'
MESSAGE_INFO = 'MESSAGE_INFO'
MESSAGES = 'MESSAGES'
METADATA = 'METADATA'
MEMORY_FRACTION = 'MEMORY_FRACTION'
MIN_PIXELS_FROM_EDGE = 'MIN_PIXELS_FROM_EDGE'
MIN_PERS_QUEUE = 'MIN_PERS_QUEUE'
MIN_PERS_QUEUE_SIZE = 'MIN_PERS_QUEUE_SIZE'
MIN_LENGTH = 'MIN_LENGTH'
MIN_TRAIN_STEPS = 'MIN_TRAIN_STEPS'

MODEL_SERVING_PROCESS = 'MODEL_SERVING_PROCESS'
MODEL_CLASSES_FILENAME = 'MODEL_CLASSES_FILENAME'
MODEL_WEIGHTS_FILENAME = 'MODEL_WEIGHTS_FILENAME'
SECOND_STAGE_MODEL_WEIGHTS_FILENAME = 'SECOND_STAGE_MODEL_WEIGHTS_FILENAME'
SECOND_STAGE_MODEL_WEIGHTS_FILENAME = 'SECOND_STAGE_MODEL_WEIGHTS_FILENAME'
CUDNN_BENCHMARK = 'CUDNN_BENCHMARK'

META_TYPE = 'META_TYPE'
MODULE = 'MODULE'
NAME = 'NAME'
NODE = 'NODE'
NO_WITNESS = 'NO_WITNESS'
NON_VIDEO_PLUGIN = 'NON_VIDEO_PLUGIN'
NR_FRAME = 'NR_FRAME'
NR_INF = 'NR_INF'
NR_INFERENCES = 'NR_INFERENCES'
NR_OBJECTS = 'NR_OBJECTS'
NR_PAYLOADS = 'NR_PAYLOADS'
NR_PERSONS = 'NR_PERSONS'
NR_PERS = 'NR_PERS'
NR_PERS_FLOOR = 'NR_PERS_FLOOR'
NR_SKIP_FRAMES = 'NR_SKIP_FRAMES'
NR_STREAMS = 'NR_STREAMS'
NR_STREAMS_DATA = 'NR_STREAMS_DATA'
OBJECT_SUBPART = 'OBJECT_SUBPART'
OBJECT_TYPE = 'OBJECT_TYPE'
OBJECTS = 'OBJECTS'
OBSTRUCTION = 'OBSTRUCTION'
OBSTRUCTION_FINGER = 'OBSTRUCTION_FINGER'
OBSTRUCTION_OTHER = 'OBSTRUCTION_OTHER'
OBSTRUCTION_RESIZE = 'OBSTRUCTION_RESIZE'
OFF = 'OFF'
ON = 'ON'
OPEN = 'OPEN'
OUTPUT = 'OUTPUT'
OUTPUT_FPS = 'OUTPUT_FPS'
OUTPUT_PATH = 'OUTPUT_PATH'
OUTPUT_PATH_PREVIEW = 'OUTPUT_PATH_PREVIEW'
OUTPUT_SCALE_FACTOR = 'OUTPUT_SCALE_FACTOR'
OUTPUT_VIDEO = 'OUTPUT_VIDEO'
OUTPUT_VIDEO_PREVIEW = 'OUTPUT_VIDEO_PREVIEW'
ORIGINAL_FRAME = 'ORIGINAL_FRAME'
PLUGINS_ON_THREADS = 'PLUGINS_ON_THREADS'
PRC_PERSON_BLUR = 'PRC_PERSON_BLUR'
PRC_VEHICLES_BLUR = 'PRC_VEHICLES_BLUR'
PERSONS_CLASSES = 'PERSONS_CLASSES'
PLUGIN_RESULTS = 'PLUGIN_RESULTS'
POINTS = 'POINTS'
PLUGINS_RESULTS = 'PLUGINS_RESULTS'
PLUGINS = 'PLUGINS'
PREDICTION_STEPS = 'PREDICTION_STEPS'
PROCESSED_COUNT = 'PROCESSED_COUNT'
PROCESSED_DATE = 'PROCESSED_DATE'
PROCESSING_RESULTS_CSV = 'PROCESSING_RESULTS_CSV'
PRC_INTERSECT = 'PRC_INTERSECT'
PRC_GOLD = 'PRC_GOLD'
PRIORITY = 'PRIORITY'
PROB_PRC = 'PROB_PRC'
PROGRESS = 'PROGRESS'
PROCESS_MEMORY = 'PROCESS_MEMORY'
PROCESS_DELAY = 'PROCESS_DELAY'
PREDICTION_STEPS = 'PREDICTION_STEPS'
RESULT = 'RESULT'
QUALITY = 'QUALITY'
QUEUE_DIRECTION = 'QUEUE_DIRECTION'
QUERY = 'QUERY'
RATIO = 'RATIO'
RECONNECTABLE = 'RECONNECTABLE'
RECORD = 'RECORD'
REGULAR = 'REGULAR'
REID = 'REID'
REIDENTIFICATIONS = 'REIDENTIFICATIONS'
REIDENTIFICATED = 'REIDENTIFICATED'
REIDENTIFICATED_TIME = 'REIDENTIFICATED_TIME'
RESULTS = 'RESULTS'
RENAME_OBJECT_TYPE = 'RENAME_OBJECT_TYPE'
RIGHT = 'RIGHT'
SAFETY_HELMET = 'SAFETY_HELMET'
SAVE_BUFFER = 'SAVE_BUFFER'
SAVE_INTERVAL = 'SAVE_INTERVAL'
SAVE_SECONDS = 'SAVE_SECONDS'
SECONDS_HEARTBEAT = 'SECONDS_HEARTBEAT'
SECONDS_READ = 'SECONDS_READ'
SECONDS_DISMISS_ALERT = 'SECONDS_DISMISS_ALERT'
SECONDS_RAISE_ALERT = 'SECONDS_RAISE_ALERT'
SECS_READ = 'SECS_READ'
SECS_RAISE_ALERT = 'SECS_RAISE_ALERT'
SECS_DISMISS_ALERT = 'SECS_DISMISS_ALERT'
SECS_NEXT_DECISION = 'SECS_NEXT_DECISION'
SELECTIVE_ZONES = 'SELECTIVE_ZONES'
SEND_EMAIL = '_H_SEND_EMAIL'
SEND_PRIORITY = 'SEND_PRIORITY'
SEND_PRIORITY_HIGH = 'SEND_PRIORITY_HIGH'
SEND_PRIORITY_MEDIUM = 'SEND_PRIORITY_MEDIUM'
SEND_PRIORITY_NORMAL = 'SEND_PRIORITY_NORMAL'
SERVER_ID = 'SERVER_ID'
SERVING_PIDS = 'SERVING_PIDS'
SHARED = 'SHARED'
SHM_OUTPUT = 'SHM_OUTPUT'
SIGNATURE = 'SIGNATURE'
SIMPLE_WITNESS = 'SIMPLE_WITNESS'
SLOW = 'SLOW'
SMALL_CIRCLE_RADIUS_PERCENT = 'SMALL_CIRCLE_RADIUS_PERCENT'
STAGE_DOWNLOAD = 'DOWNLOAD'
STAGE_PROCESS = 'PROCESS'
STAGE_PING = 'PING'
STAGE_SEARCH_WORKERS = 'SEARCH_WORKERS'
STANDING_STILL = 'STANDING_STILL'
STANDING_STILL_TIME = 'STANDING_STILL_TIME'
START = 'START'
START_TIME = 'START_TIME'
STATE = 'STATE'
STATUS = 'STATUS'
STATUS_CHANGED = 'STATUS_CHANGED'
STEP = 'STEP'
STEP_MINUTES = 'STEP_MINUTES'
STILL_SECONDS = 'STILL_SECONDS'
STILL_TIME = 'STILL_TIME'
STREAM = 'STREAM'
STREAM_ID = 'STREAM_ID'
STREAM_CONFIG_METADATA = 'STREAM_CONFIG_METADATA'
STREAM_DEBUG_DATA = 'STREAM_DEBUG_DATA'
STREAM_IMG = 'STREAM_IMG'
STREAM_INFER = 'STREAM_INFER'
STREAM_INFER_INTERSECT = 'STREAM_INFER_INTERSECT'
STREAM_MESSAGES = 'STREAM_MESSAGES'
STREAM_PLUGINS_RESULTS = 'STREAM_PLUGINS_RESULTS'
STREAM_STEP_RESULTS = 'STREAM_STEP_RESULTS'
STREAMS = 'STREAMS'
STREAMS_KEY = 'STREAMS_KEY'
STRIDE = 'STRIDE'
SYSTEM_TIME = 'SYSTEM_TIME'
RESERVED_MEM = 'RESERVED_MEM'
TARGET_SUBPART = 'TARGET_SUBPART'
THR_MASK = 'THR_MASK'
THREADS_PREFIX = 'S_'
THRESHOLD = 'THRESHOLD'
TIME = 'TIME'
TOPIC_DEVICE_SPECIFIC = 'TOPIC_DEVICE_SPECIFIC'
TOPIC_SEND = 'TOPIC_SEND'
TOPIC_RECV = 'TOPIC_RECV'
TRACKING_ALIVE_TIME = 'TRACKING_ALIVE_TIME'
TRACKING_MODE = 'TRACKING_MODE'
TRACKING_OBJECTS = 'TRACKING_OBJECTS'
PAYLOADS_CHANNEL = 'PAYLOADS_CHANNEL'
CONFIG_CHANNEL = 'CONFIG_CHANNEL'
TOPIC_SEND_DEVICE_SPECIFIC = 'TOPIC_SEND_DEVICE_SPECIFIC'
TOPIC_RECV_DEVICE_SPECIFIC = 'TOPIC_RECV_DEVICE_SPECIFIC'
TOTAL_MEM = 'TOTAL_MEM'
TOTAL_DISK = 'TOTAL_DISK'
TOTAL_NR_PERS = 'TOTAL_NR_PERS'
TOTAL_MESSAGES = 'TOTAL_MESSAGES'
TIME_IN_TARGET = 'TIME_IN_TARGET'
TRACK_ID = 'TRACK_ID'
TRACKING_STARTED_AT = 'TRACKING_STARTED_AT'
TRAIN_PERIODS = 'TRAIN_PERIODS'
TRAIN_HIST = 'TRAIN_HIST'
TRAIN_STEPS = 'TRAIN_STEPS'
TRANSCODER_H = 'TRANSCODER_H'
TRANSCODER_W = 'TRANSCODER_W'
TURNSTILE_SECONDS_INTERVAL = 'TURNSTILE_SECONDS_INTERVAL'
QOS = 'QOS'
UNKNOWN = 'UNKNOWN'
UPLOAD_MOVIE = 'UPLOAD_MOVIE'
UPLOAD_ZIP = 'UPLOAD_ZIP'
UPTIME = 'UPTIME'
URL = 'URL'
URL_CONFIG = 'URL_CONFIG'
URL_DOWNLOAD = 'URL_DOWNLOAD'
URL_DETECTIONS = 'URL_DETECTIONS'
URL_MODELS_ZOO = 'URL_MODELS_ZOO'
URLS = 'URLS'
USE = 'USE'
VALUE = 'VALUE'
VANISHING = 'VANISHING'
VEHICLE_CLASSES = 'VEHICLE_CLASSES'
VIDEO_FILE = 'VIDEO_FILE'
VIDEO_STREAM = 'VIDEO_STREAM'
VERSION = 'VERSION'
WEAPON_ASSAILANT = 'WEAPON_ASSAILANT'
WIDTH_PRC = 'WIDTH_PRC'
WHITELIST = 'WHITELIST'
WHITELISTED = 'WHITELISTED'
WHITELISTED_TIME = 'WHITELISTED_TIME'
TIMERS = 'TIMERS'
TYPE = 'TYPE'
THR_ANGLE_ALERT = 'THR_ANGLE_ALERT'
TIMESTAMP = 'TIMESTAMP'
TLBR_POS = 'TLBR_POS'
TLBR_POS_TRACK = 'TLBR_POS_TRACK'
TOP = 'TOP'
TOP_N_SOFTMAX = 'TOP_N_SOFTMAX'
TURNSTILE_SEC = 'TURNSTILE_SEC'
ZONE = 'ZONE'
ZONES = 'ZONES'
ZONE_NR_PERS = 'ZONE_NR_PERS'
ZIP_SIZE = 'ZIP_SIZE'
WRONG = 'WRONG'
WARMUP_FRAMES = 'WARMUP_FRAMES'
WARMUP_ITERATIONS = 'WARMUP_ITERATIONS'


NO_MESSAGE = "STATUS UNCHANGED"
INFO_MSG = "INFO_MSG"

# LISTS
LIST_MOVIE = ['.avi', '.mp4', '.mkv']
LIST_IMAGE = ['.png', '.jpg', '.jpeg', '.bmp']
LIST_PKL = ['.pkl', '.pk']


PLUGIN_INSTANCE_PARAMETER_LIST = 'PLUGIN_INSTANCE_PARAMETER_LIST'


# DRAWING
# BGRs
DARK_BLUE = (109, 45, 11)
BLUE = (255, 0, 0)
LIGHT_BLUE = (230, 200, 100)
ORANGE = (19, 149, 250)
WHITE = (255, 255, 255)
BLACK = (0, 0, 0)
YELLOW = (98, 222, 250)
RED = (100, 70, 248)
GREEN = (20, 180, 20)
DARK_GREEN = (20, 120, 20)
PALE_GREEN = (60, 230, 170)
DEEP_RED = (20, 20, 170)
BROWN = (42, 42, 165)
PURPLE = (211, 0, 148)


# OTHER
LIGHT_BLUR = (254, 255, 129)


LOW_DRAWING = 'LOW_DRAWING'

# END DRAWING


# DetectionClasses
CAR = 'car'
FACE = 'face'
PERSON = 'person'
PERSOANA = 'persoana'


BLUR_PERSON_DEFAULT = 'default'
BLUR_PERSON_ADAPTIVE = 'adaptive'

RECTANGLE_TYPE_DEFAULT = 'default'
RECTANGLE_TYPE_CORNERS = 'corners'

# #CAMERA QUALITY CHECK
# CQC_COMPRESS = 'COMPRESS'
# CQC_DIRT = 'DIRT'
# CQC_EMP = 'EMP'
# CQC_FOG = 'FOG'
# CQC_GOOD = 'GOOD'
# CQC_ICE = 'ICE'
# CQC_OBSCURED = 'OBSCURED'
# CQC_SNOW = 'SNOW'
# CQC_WATER = 'WATER'

# GRID HEATMAP
GRID_X_COUNT = 'GRID_X_COUNT'
GRID_Y_COUNT = 'GRID_Y_COUNT'
REF_AREA_TL = 'REF_AREA_TL'
REF_AREA_TR = 'REF_AREA_TR'
REF_AREA_BL = 'REF_AREA_BL'
REF_AREA_BR = 'REF_AREA_BR'
TRANSFORM_WHOLE = 'TRANSFORM_WHOLE'
HEAT_RANGE = 'HEAT_RANGE'
HEATMAP_RECORD_INTERVAL = 'HEATMAP_RECORD_INTERVAL'
HEATMAP_PERSON_LOW = 'HEATMAP_PERSON_LOW'
HEATMAP_PERSON_LOW_DEFAULT = 0.1

# LOCATIONS
SENSORS_LOCATION = 'naeural_core.utils.sensors'

CAP_ACTUAL_DPS = 'CAP_ACTUAL_DPS'
CAP_CACHE_MAX = 'CAP_CACHE_MAX'
CAP_CACHE_LEN = 'CAP_CACHE_LEN'
CAP_LIVE = 'CAP_LIVE'
CAP_RECONN = 'CAP_RECONNECTABLE'


# DATA ACQUISITION, PAYLOADS & COMM RESOLUTIONS & other stuff
CAPTURE_MANAGER = 'CAPM'

NR_CAPTURES = 'NR_CAPTURES'

TRANSCODER_FAILS = 200
MAIN_LOOP_RESOLUTION = 20

DEFAULT_SECONDS_HEARTBEAT = 10
MAX_SECONDS_HEARTBEAT = 10

CAP_RESOLUTION_DEFAULT = 4
# we have queue protection so we do not need large queue
CAP_RESOLUTION_NON_LIVE_DEFAULT = 32

INIT_DATA = 'INIT_DATA'
CAP_RESOLUTION = 'CAP_RESOLUTION'

CAPTURES_PER_SEC = 'CAPTURES_PER_SEC'
STREAM_TIME = 'STREAM_TIME'

CAPTURE_STATS_DISPLAY = 'CAPTURE_STATS_DISPLAY'
CAPTURE_STATS_DISPLAY_DEFAULT = 120  # in seconds

OBSOLETE_FLAGS = ['CAP_READ_DELAY',]

PLUGIN_EXCEPTION_DELAY = 10


COMM_SEND_BUFFER = 1000
COMM_RECV_BUFFER = 100
COMM_SECS_SHOW_INFO = 60


# Payload constants
IDLE_PLUGIN_ITERS = CAP_RESOLUTION_DEFAULT * 20
PLUGIN_REAL_RESOLUTION = 'PLUGIN_REAL_RESOLUTION'
PLUGIN_LOOP_RESOLUTION = 'PLUGIN_LOOP_RESOLUTION'


RESIZE_H = 'RESIZE_H'
RESIZE_W = 'RESIZE_W'
RESIZE_WITNESS = 'RESIZE_WITNESS'
FRAME_CROP = 'FRAME_CROP'
CAPTURE_CROP = 'CAPTURE_CROP'  # alias for FRAME_CROP

FORCED_DELAY = 5

# END DATA AREA


# INFERENCE ENGINE

PICKED_INPUT = 'PICKED_INPUT'
RUNS_ON_EMPTY_INPUT = 'RUNS_ON_EMPTY_INPUT'
MODEL_INSTANCE_ID = 'MODEL_INSTANCE_ID'


TIMER_FILTER_RESULTS = 'filter_results'
TIMER_LOAD_GRAPH = 'load_graph'
TIMER_PREDICT = 'pred'
TIMER_PREDICT_BATCH = 'pred_batch'
TIMER_PREPARE_PAYLOAD = 'prep_payload'
TIMER_PREPARE_RESULTS = 'prep_results'
TIMER_PRE_PROCESS = 'pre_proc'
TIMER_POST_PROCESS = 'post_proc'
TIMER_PREPROCESS_INPUTS = 'pre_proc_inp'
TIMER_POST_PROCESS_INFERENCE = 'post_proc_inf'
TIMER_POST_PROCESS_OUTPUTS = 'post_proc_out'
TIMER_RUN_INFERENCE = 'run_inf'
TIMER_PACK_RESULTS = 'pack_res'
TIMER_SESSION_RUN = 'sess_run'

BATCH_STRATEGY_DEFAULT = 'DEFAULT'
BATCH_STRATEGY_MOST_COMMON_SHAPE = 'MOST_COMMON_SHAPE'
BATCH_STRATEGY_PER_SHAPE = 'PER_SHAPE'
CLASSES = 'CLASSES'
CONFIG = 'CONFIG'
CONFIG_INFERENCE = 'CONFIG_INFERENCE'
CONFIG_PLUGINS = 'CONFIG_PLUGINS'
CONFIG_STREAMS = 'CONFIG_STREAMS'
ERROR = 'ERROR'
GRAPH = 'GRAPH'
IMG = 'IMG'
IMG_H = 'IMG_H'
IMG_W = 'IMG_W'
IOU = 'IOU'
INPUT_TENSORS = 'INPUT_TENSORS'
INFERENCES = 'INFERENCES'
IOU_THRESHOLD = 'IOU_THRESHOLD'
MODEL_THRESHOLD = 'MODEL_THRESHOLD'
METADATA = 'METADATA'
NMS = 'NMS'
OUTPUT_TENSORS = 'OUTPUT_TENSORS'
STYLE_IMAGE = 'STYLE_IMAGE'
SYSTEM_TIME = 'SYSTEM_TIME'
VER = 'VER'


# default confidence threshold should be high in most cases OR
# we can use AlertHelper with percentages and mean of 0.51 or higher


# TORCH constants
WARMUP_BATCH = 'WARMUP_BATCH'
USE_AMP = 'USE_AMP'
URL_CLASS_NAMES = 'URL_CLASS_NAMES'
MODEL_NAME = 'MODEL_NAME'
NMS_CONF_THR = 'NMS_CONF_THR'
NMS_IOU_THR = 'NMS_IOU_THR'
NMS_MAX_DET = 'NMS_MAX_DET'
YAML_PATH = 'YAML_PATH'
DEFAULT_BATCH = 'DEFAULT_BATCH'
URL_BY_BATCH = 'URL_BY_BATCH'
DEFAULT_DEVICE = 'DEFAULT_DEVICE'
USE_FP16 = 'USE_FP16'
GPU_PREPROCESS = 'GPU_PREPROCESS'
MAX_BATCH_FIRST_STAGE = 'MAX_BATCH_FIRST_STAGE'
MAX_BATCH_SECOND_STAGE = 'MAX_BATCH_SECOND_STAGE'


# INFERENCE
TIMER_CALCULATE_BENCHMARK_TIME = 'calculate_benchmark_time'
TIMER_INFER = 'timer_infer'
TIMER_LOAD_GRAPH = 'load_graph'
TIMER_POSTPROCESS_BOXES = 'postprocess_boxes'
TIMER_POSTPROCESS_IMAGES = 'postprocess_images'
TIMER_PREDICT = 'predict'
TIMER_PREDICT_BATCH = 'predict_batch'
TIMER_PREPARE_PAYLOAD = 'prepare_payload'
TIMER_PREPARE_RESULTS = 'prepare_results'
TIMER_PREPROCESS_IMAGES = 'preprocess_images'
TIMER_RUN_INFERENCE = 'run_inference'
TIMER_SESSION_RUN = 'session_run'

BATCH_STRATEGY_DEFAULT = 'DEFAULT'
BATCH_STRATEGY_MOST_COMMON_SHAPE = 'MOST_COMMON_SHAPE'
BATCH_STRATEGY_PER_SHAPE = 'PER_SHAPE'

CONFIG = 'CONFIG'
CONFIG_INFERENCE = 'CONFIG_INFERENCE'
CONFIG_PLUGINS = 'CONFIG_PLUGINS'
CONFIG_STREAMS = 'CONFIG_STREAMS'

GRAPH = 'GRAPH'
GRAPH_TYPE_PB = 'pb'
GRAPH_TYPE_ONNX = 'onnx'
GRAPH_TYPE_TENSORFLOW_TRT = 'tensorflow_trt'
GRAPH_TYPE_PYTORCH_TRT = 'pytorch_trt'
GRAPH_TYPE_KERAS = 'keras'
GRAPH_TYPE_PYTORCH = 'pytorch'

GRAPH_TASK_CLASSIFICATION = 'classification'
GRAPH_TASK_DETECTION = 'detection'

FRAMEWORK_TENSORFLOW = 'tensorflow'
FRAMEWORK_PYTORCH = 'pytorch'

INFERENCE_GRAPHS = 'INFERENCE_GRAPHS'
INFERENCE_META_GRAPHS = 'INFERENCE_META_GRAPHS'
INFERENCE_ON_FULL_IMAGE = 'INFERENCE_ON_FULL_IMAGE'
IOU_THRESHOLD = 'IOU_THRESHOLD'

LINKED_INSTANCES = 'LINKED_INSTANCES'
LINKED_SERVER = 'LINKED_SERVER'


LOCATIONS = 'LOCATIONS'

MODEL_THRESHOLD = 'MODEL_THRESHOLD'

PLUGINS = 'PLUGINS'
SIGNATURE = 'SIGNATURE'
USE_GRAPHS = 'USE_GRAPHS'


ALIMENTARE_NEPERMISA = 'ALIMENTARE_NEPERMISA'
CAMERA_QUALITY_CHECK = 'CAMERA_QUALITY_CHECK'
COVID_MASK = 'COVID_MASK'
EFF_DET0 = 'EFF_DET0'
FIRE_SMOKE = 'FIRE_SMOKE'
FACE_DETECTION = 'FACE_DETECTION'
GASTRO = 'GASTRO'
LP_DETECTION = 'LP_DETECTION'
LPDV2 = 'LPDV2'
LPR = 'LPR'
MERCHANDISER = 'MERCHANDISER'
OMV_EMPLOYEE = 'OMV_EMPLOYEE'
TF_YOLO = 'TF_YOLO'
TF_ODAPI1 = 'TF_ODAPI1'
TF_ODAPI1_OIDV4 = 'TF_ODAPI1_OIDV4'
TF_ODAPI1_TRAFFICSIGNS = 'TF_ODAPI1_TRAFFICSIGNS'
TF_ODAPI2 = 'TF_ODAPI2'

LST_VEHICLES = ['car', 'truck', 'bus']

WEEKDAYS = [
  'MONDAY',
  'TUESDAY',
  'WEDNESDAY',
  'THURSDAY',
  'FRIDAY',
  'SATURDAY',
  'SUNDAY'
]


class ThAnchor:
  ANCHOR_UPDATED_AT = "ANCHOR_UPDATED_AT"
  ANCHOR_METADATA = "ANCHOR_METADATA"
  ANCHOR_IMAGES = "ANCHOR_IMAGES"
  ANCHOR_RELOAD_TIME = "ANCHOR_RELOAD_TIME"

  PREDICTIONS = "PREDICTIONS"


class DatasetBuilder:
  FILE = 'FILE'
  LABEL_TEMPLATE_TYPE = 'LABEL_TEMPLATE_TYPE'
  LABEL_FILE_TEMPLATE = 'LABEL_FILE_TEMPLATE'
  LABEL_EXTENSION = 'LABEL_EXTENSION'
  X_EXTENSION = 'X_EXTENSION'
  DATASET_BUILDER = 'DATASET_BUILDER'
  DATASET_NAME = 'DATASET_NAME'
  VALID_FILE_TYPES = ['FILE', None]
  INFERENCE_MAPPING = 'INFERENCE_MAPPING'


class ForceStopException(Exception):
  pass

CONFIG_STARTUP_MANDATORY_KEYS = [
  EE_ID, 
  CONFIG_STARTUP_v2.SECURED,
  "MAIN_LOOP_RESOLUTION",
  SECONDS_HEARTBEAT,
  CONFIG_STARTUP_v2.K_CONFIG_RETRIEVE,
  PLUGINS_ON_THREADS,
]

# former constants and config_startup.json HB_CONTAINS_PIPELINES and HB_CONTAINS_ACTIVE_PLUGINS 
# are now exclusively used as environment variables
HB_CONTAINS_PIPELINES_ENV_KEY = 'EE_HB_CONTAINS_PIPELINES'
HB_CONTAINS_ACTIVE_PLUGINS_ENV_KEY = 'EE_HB_CONTAINS_ACTIVE_PLUGINS'

try:
  from constants import *
except:
  pass


from collections import OrderedDict


class _DebugInfoMixin(object):
  def __init__(self):
    self._debug_info = []
    super(_DebugInfoMixin, self).__init__()
    return

  @staticmethod
  def _format_object_for_debug_info(obj, depth=0):
    if isinstance(obj, dict):
      if len(obj) == 0:
        msg = '{}'
      else:
        content = ' '.join(['{}:{}'.format(k, _DebugInfoMixin._format_object_for_debug_info(v, depth + 1))
                            for k, v in obj.items()])
        if depth > 0:
          msg = '{{{}}}'.format(content)
        else:
          msg = content
    elif isinstance(obj, list):
      # maybe compress the list?
      # but simply run the formatter on each of the objects contained
      if len(obj) == 0:
        msg = '[]'
      else:
        content = ' '.join([_DebugInfoMixin._format_object_for_debug_info(v, depth + 1) for v in obj])
        if depth > 0:
          msg = content
          # msg = '[{}]'.format(content) TODO: uncomment ar review time
        else:
          msg = content
    elif isinstance(obj, float):
      msg = f"{obj:.2f}"
    elif isinstance(obj, tuple):
      msg = ""
      for x in obj:
        msg += _DebugInfoMixin._format_object_for_debug_info(x, depth + 1)
    else:
      msg = str(obj)

    return msg

  def reset_debug_info(self):
    self._debug_info = []
    return

  def add_debug_info(self, value, key: str = None):
    """
    Add debug info to the witness. The information will be stored in a new line.

    Parameters
    ----------
    value : Any
        The info to be shown
    key : str
        The key that identifies this debug line
    """
    self._debug_info.append({
      'key': key,
      'value': value,
    })

    return

  def _draw_debug_info_on_witness(self, img_witness):
    img_sizes = [400, 700, 1000]
    font_sizes = {
      0: 0.2,
      1: 0.3,
      2: 0.4,
      3: 0.7
    }

    # This list will contain all the lines that will be drawn on the witness
    debug_info_lines = []

    # compute alerter data strings; TODO: check if default alerter is used
    any_alert = False
    for alerter_name in self.alerters_names:
      last_change_time = self.get_alerter(alerter_name).get_time_from_change()
      _is_default_alerter = (alerter_name == "default")
      _show_default_alerter = (_is_default_alerter and self.cfg_debug_show_default_alerter)
      _is_alert = self.alerter_is_alert(alerter_name)
      _not_changed_recently = last_change_time == -1 or last_change_time > self.cfg_debug_info_active_alerter_time
      _skip_alerter = not _show_default_alerter and (not _is_alert and _not_changed_recently)

      if _skip_alerter:
        # skip showing this alerter if no data has been added to it in the last
        # `DEBUG_INFO_ACTIVE_ALERTER_TIME` seconds
        continue

      is_alert = self.alerter_is_alert(alerter_name)
      txt_alert = '[{}] Alert: {}, {}'.format(
          alerter_name[:3], is_alert, self.get_alerter_status(alerter_name)
      )
      debug_info_lines.append(txt_alert)

      any_alert = any_alert or is_alert

    # encapsulating in try-catch so that we can dump the error and protect the OrderedDict
    # compute debug data strings
    for dct in self._debug_info:
      k, v = dct['key'], dct['value']

      txt_debug = ""
      if k is not None:
        txt_debug += '[{}] '.format(k)
      txt_debug += '{}'.format(self._format_object_for_debug_info(v))

      debug_info_lines.append(txt_debug)

    # draw debug info
    img_size = img_witness.shape[1]
    font_size = font_sizes[sum([img_size > s for s in img_sizes])]  # 0.7 if img_size > 1000 else 0.4
    font_thickness = 2 if img_size > 1000 else 1
    img_witness = self._painter.draw_text_outer_image(
        image=img_witness,
        text=debug_info_lines,
        font_size=font_size,
        thickness=font_thickness,
        color=self.consts.DEEP_RED if any_alert else self.consts.WHITE,
        location="top",
    )

    # reset debug info for next iteration
    self.reset_debug_info()

    return img_witness



"""
  TODO:
    Add GPU SYNC !
"""
from naeural_core import constants as ct
from naeural_core.serving.base import UnifiedFirstStage as ParentServingProcess
from naeural_core.local_libraries.nn.th.utils import th_resize_with_pad

_CONFIG = {
  **ParentServingProcess.CONFIG,

  'IMAGE_HW': (840, 840),  # The model is size-invariant, but is trained on this size ( 1080, 1920),#
  'MAX_BATCH_FIRST_STAGE': 5,
  "MODEL_WEIGHTS_FILENAME": "20230723_RET_FACE_RES_bs1_nms.ths",
  "URL": "minio:retina_face/20230723_RET_FACE_RES_bs1_nms.ths",


  'VALIDATION_RULES': {
    **ParentServingProcess.CONFIG['VALIDATION_RULES'],
  },
}

__VER__ = '0.1.0.0'


class ThRface(ParentServingProcess):
  CONFIG = _CONFIG

  def __init__(self, **kwargs):
    ver = kwargs.pop('version', None)
    if ver is None:
      ver = __VER__

    # this must be correctly configured by each individual serving process
    self._has_second_stage_classifier = False
    self.resized_input_images = None
    super(ThRface, self).__init__(version=ver, **kwargs)
    return

  def _startup(self):
    super(ThRface, self)._startup()
    # self.norm_sub_tensor = th.Tensor([123, 117, 104]).reshape((3,1,1)).to(self.dev)
    self.norm_sub_tensor = self.th.Tensor([104, 117, 123]).reshape((3, 1, 1)).to(self.dev)

  @property
  def has_second_stage_classifier(self):
    return self._has_second_stage_classifier and self.server_name == self.predict_server_name

  def _get_model(self, config):
    model = self._prepare_ts_model(fn_model=config, post_process_classes=False)
    return model

  def _pre_process_images(self, images):
    lst_original_shapes = []

    # Images must be BGR
    images = [self.cv2.cvtColor(x, self.cv2.COLOR_RGB2BGR) for x in images]

    # run GPU based pre-processing
    self._start_timer('resize')
    prep_inputs, lst_original_shapes = th_resize_with_pad(
      img=images,
      h=self.cfg_input_size[0],
      w=self.cfg_input_size[1],
      device=self.dev,
      normalize=True,
      sub_val=self.norm_sub_tensor,
      div_val=1,
      half=self.cfg_fp16
    )
    self._stop_timer('resize')

    self._lst_original_shapes = lst_original_shapes
    return prep_inputs

  def _aggregate_batch_predict(self, lst_results):
    th_preds_batch = self.th.cat([x[0] for x in lst_results])
    th_preds_n_det = self.th.cat([x[1] for x in lst_results])

    return th_preds_batch, th_preds_n_det

  def _second_stage_process(self, th_preds, th_inputs=None, **kwargs):
    th_pred_nms, th_n_dets = th_preds
    self.log.start_timer('slice_det')
    pred_nms = [x[:th_n_dets[i]] for i, x in enumerate(th_pred_nms)]
    self.log.stop_timer('slice_det')

    pred_second = None
    if self.has_second_stage_classifier:
      self.log.start_timer('2nd_stage_clf')
      pred_second = self._second_stage_classifier(pred_nms, th_inputs)
      self.log.stop_timer('2nd_stage_clf')
    # endif

    self.log.start_timer('yolo_to_cpu')
    pred_nms_cpu = [x.cpu().numpy() for x in pred_nms]
    self.log.stop_timer('yolo_to_cpu')

    return pred_nms_cpu, pred_second

  def _post_process(self, preds):
    pred_nms_cpu, _ = preds
    nr_images = len(self._lst_original_shapes)
    lst_results = []
    for i in range(nr_images):
      # now we have each individual image and we generate all objects
      # what we need to do is to match `second_preds` to image id & then
      # match second clf with each box
      np_pred_nms_cpu = pred_nms_cpu[i]
      original_shape = self._lst_original_shapes[i]

      if np_pred_nms_cpu.shape[0] != 0:
        denorm_array = self.np.array(
          [self.cfg_input_size[1], self.cfg_input_size[0]] * 2 +
          [1] +
          [self.cfg_input_size[1], self.cfg_input_size[0]] * 5
        )
        np_pred_nms_cpu_denormed = np_pred_nms_cpu * denorm_array
        np_pred_nms_cpu[:, :4] = self.scale_coords(
          img1_shape=self.cfg_input_size,
          coords=np_pred_nms_cpu_denormed[:, :4],
          img0_shape=original_shape,
        ).round()

        np_pred_nms_cpu[:, 5:] = self.scale_coords(
          img1_shape=self.cfg_input_size,
          coords=np_pred_nms_cpu_denormed[:, 5:],
          img0_shape=original_shape,
        ).round()

      ####

      lst_inf = []
      for det in np_pred_nms_cpu:
        det = [float(x) for x in det]
        # order is [left, top, right, bottom, proba] =>
        # [L, T, R, B, P, P1x, P1y, P2x, P2y, P3x, P3y, P4x, P4y, P5x, P5y]
        # where Pix and Piy are coordinates for Pi and
        # Pi is from [right ear, left eat, nose, right mouth corner, left mouth corner]
        L, T, R, B, P = det[:5]  # order is [left, top, right, bottom, proba]
        kpts = det[5:]
        dct_obj = {
          ct.TLBR_POS: [int(T), int(L), int(B), int(R)],
          ct.PROB_PRC: round(float(P), 2),
          ct.TYPE: "FACE",
          "KEYPOINTS": [(int(kpts[i]), int(kpts[i + 1])) for i in range(0, len(kpts), 2)]
        }
        lst_inf.append(dct_obj)
      lst_results.append(lst_inf)
    return lst_results



import cv2

def read_qr(np_rgb_img, debug=False):
  USE_TRICK = False
  qr_det = cv2.QRCodeDetector()
  if USE_TRICK:
    # now we perform a trick - convert to BGR then convert from ... RGB to GRAY
    np_bgr_img = np_rgb_img[:,:,::-1] 
    np_gray_img = cv2.cvtColor(np_bgr_img, cv2.COLOR_RGB2GRAY)
  else:
    np_gray_img = np_rgb_img.mean(axis=-1, keepdims=True).astype('uint8')
  if debug:
    retval, points = qr_det.detect(np_gray_img)
    print("QR DEBUG: {}:'{}'".format(retval, points))
  data, bbox, qr_mat = qr_det.detectAndDecode(np_gray_img)
  top, left, bottom, right, np_slice = None, None, None, None, None
  if bbox is not None:  
    top = int(bbox[:,:,1].min())
    left = int(bbox[:,:,0].min())
    bottom = int(bbox[:,:,1].max())
    right = int(bbox[:,:,0].max())
    
    np_slice = np_gray_img[top:bottom, left:right].copy()
  
    return data, (top, left, bottom, right), bbox, qr_mat, np_slice
  return


import cv2
import numpy as np
import tensorflow as tf

from tqdm import tqdm
from naeural_core import Logger

if __name__ == '__main__':
  cfg_file = 'main_config.txt'
  log = Logger(
    lib_name='XP_BMW', 
    config_file=cfg_file, 
    max_lines=1000, 
    TF_KERAS=False
    )

  TARGET_H = 1520
  TARGET_W = 2688

  img_bgr1 = cv2.imread(r'D:/vlcsnap-2021-09-01-10h01m50s069.png')
  img_bgr2 = cv2.imread(r'D:/vlcsnap-2021-09-01-10h01m50s069.png')
  
  img_rgb1 = img_bgr1[:,:,::-1]
  img_rgb2 = img_bgr2[:,:,::-1]
  
  np_imgs_bgr = np.array([img_bgr1, img_bgr2, img_bgr2])
  np_imgs_rgb = np.array([img_rgb1, img_rgb2, img_rgb2])
  # np_imgs_rgb = img_rgb1
  
  for _ in tqdm(range(20)):
    log.start_timer('log_center_image2')    
    l = []
    for img in np_imgs_bgr:
      img_center = log.center_image2(
        np_src=img, 
        target_h=TARGET_H, 
        target_w=TARGET_W
        )
      l.append(img_center)
    arr = np.array(l)
    tf_imgs = tf.constant(arr, dtype=np.uint8) + 0
    log.stop_timer('log_center_image2')    
    
    log.start_timer('tf.image.resize')
    tf_img_res = tf.image.resize(
      images=np_imgs_rgb, 
      size=(TARGET_H, TARGET_W), 
      preserve_aspect_ratio=True
      )
    # np_img_res = tf_img_res.numpy().astype(np.uint8)
    log.stop_timer('tf.image.resize')
    
    log.start_timer('tf.image.resize_with_pad')
    tf_img_res_pad = tf.image.resize_with_pad(
      image=np_imgs_rgb, 
      target_height=TARGET_H, 
      target_width=TARGET_W
      )
    # np_img_res_pad = tf_img_res_pad.numpy().astype(np.uint8)
    log.stop_timer('tf.image.resize_with_pad')

  if False:  
    cv2.imshow('Original', img_bgr)
    cv2.imshow('Log center_image2', img_center)
    cv2.imshow('tf.image.resize', tf_img_res.numpy()[:,:,::-1].astype(np.uint8))
    cv2.imshow('tf.image.resize_with_pad', tf_img_res_pad.numpy()[:,:,::-1].astype(np.uint8))
    cv2.imshow('img_rgb', img_rgb)
    cv2.waitKey(0)
    cv2.destroyAllWindows()


  log.show_timers()


import os
from naeural_core import Logger
from naeural_core.config.base import BaseConfigRetrievingPlugin

class _LocalConstants:
  PATH = 'PATH'
  IS_RELATIVE_PATH = 'IS_RELATIVE_PATH'

_CONFIG = {
  **BaseConfigRetrievingPlugin.CONFIG,
  'VALIDATION_RULES': {
    **BaseConfigRetrievingPlugin.CONFIG['VALIDATION_RULES'],
  },
}

class LocalConfigRetriever(BaseConfigRetrievingPlugin):
  CONFIG = _CONFIG
  def __init__(self, log : Logger, **kwargs):
    super(LocalConfigRetriever, self).__init__(log=log, prefix_log='[LocalCR]', **kwargs)
    return
  
  def _connect(self, **kwargs):
    self.P("  'Fake' connect - locally accessing file...")
    return
  

  def __create_full_path(self, endpoint):
    if isinstance(endpoint, dict):
      path, is_relative_path = endpoint['PATH'], endpoint['IS_RELATIVE_PATH']
    elif isinstance(endpoint, str):
      path = endpoint
      is_relative_path = False
    else:
      raise ValueError('Unknown endpoint type. Supported types: dict, str')
    if is_relative_path:
      path = os.path.join(self.log.get_base_folder(), path)
    return path

  def _get_app_configuration(self, endpoint):
    path = self.__create_full_path(endpoint)
    config_app = self.log.load_json(path, folder=None, verbose=True)
    return config_app

  def _get_streams_configurations(self, endpoint):
    assert isinstance(endpoint, dict), "`endpoint` must be dict, received({}): {}".format(
      type(endpoint), endpoint,
    )
    path = self.__create_full_path(endpoint)
    lst_streams = [self.log.load_json(os.path.join(path, fn), folder=None, verbose=False) for fn in os.listdir(path)]
    return lst_streams



import numpy as np
from collections import deque
from time import time

__VER__ = '1.3.0'

MAX_QUEUE_SHOW = 8

class AlertHelper:
  # atomic & thread-safe class, no logging inside it so no subclassing of DecentrAIObject
  def __init__(self, 
               name,
               values_count, 
               raise_confirmation_time, 
               lower_confirmation_time,
               raise_alert_value=0.5, 
               lower_alert_value=0.4,
               alert_mode='mean',
               alert_mode_lower='max',
               reduce_value=False,
               reduce_threshold=0.5,
               show_version=True,
               ):
    """
    Create a AlertHelper object
    
    The pseudo-code is as follows:
      
      1. add data to data_train
      2. if eval(data_train) >= raise_threshold and status is not-alerted:
        2.1. raise alert
      3. else if status is alerted and eval(data_train) < lower_threshold
        3.1 lower alert
      
      

    Parameters
    ----------
    
    values_count : int
      the number of observations that are required for a state change evaluation.
    
    raise_confirmation_time : int
      seconds that need to pass for the confirmation of a state change. if during this time
      the queue buffer is overrun then it will be extended for accurate evaluation. 
      OBSERVATION: 
        please do not make any confusions with acquisition time. This parameter 
        controls the finite state machine behaviour from the moment the available date 
        in queue is eligible for a state change
        
    lower_confirmation_time: int
      seconds that need to pass in order to lower alert from 1 to 0.
      Observation:  it is preferable to have longer wait time as well as smaller threshold when 
                    lowering alarms

    
    raise_alert_value : int, optional
      the result that triggers the True state. Depends on 'alert mode. The default is 0.5.
      the transition is triggered at first eval >= raise_alert_value
      
          
    lower_alert_value: int, optional
      the results that enables transition from alert mode to non-alert mode. Default 0.4.
      as long as the eval is >= lower_alert_value the raised status will be kept and no transition will
      be triggered      
      Observation:  it is preferable to have longer wait time as well as smaller threshold when 
                    lowering alarms. For example it is better that if the raise is done with a 
                    0.5 threshold and the current eval is around that value to have a lower threshold 
                    that will ensure a "certainty" of the alert lowering - eg I will raise the allarm when
                    I have 0.5 (50%) eval but if the alarm is already raised I will not lower it unless we have
                    below a 0.4 (40%) evaluation
          
    reduce_value: bool, optional,
      this flag reduces the added observation to 0 or 1 for non zero values, default `False`
    
    alert_mode : str, optional
      The alert mode - 'mean' will compute mean of values and compare it with `alert_value` 
      while other 'sum' will generate a comparison of the `sum(queue)` with `alert_value`. 
      The default is 'mean'.

    alert_mode_lower : str, optional
      The same as alert_mode with the exception that this will be applied when trying to lower an alert.
      The default is 'max'.

    Returns
    -------
    None.

    """
    self.name = name + '_AH'
    self.__version__ = __VER__
    self._show_version = show_version
    self.version = self.__version__
    self._eval_func, self._eval_method = self.get_eval_func_method(alert_mode=alert_mode)
    self._eval_func_lower, self._eval_method_lower = self.get_eval_func_method(alert_mode=alert_mode_lower)
    self._queue_size = values_count
    self._raise_wait_time = raise_confirmation_time
    self._lower_wait_time = lower_confirmation_time
    self._raise_alert_value = raise_alert_value
    self._lower_alert_value = lower_alert_value
    self._alert_mode = alert_mode
    self._alert_mode_lower = alert_mode_lower
    self._reduce_value = reduce_value
    self._reduce_threshold = reduce_threshold
    self._eval_queue = None
    self._times_queue = None
    self._change_time = None
    self._eval_queue_raw = None
    self._eval_times = None
    self._changing_state = None
    
    self._last_raise_timestamp = None
    self._last_lower_timestamp = None

    self.reset()
    return

  def get_eval_func_method(self, alert_mode):
    alert_mode = alert_mode.lower()
    if alert_mode == 'sum':
      return np.sum, 'S'
    elif alert_mode == 'mean':
      return np.mean, 'A'
    elif alert_mode == 'median':
      return np.median, 'M'
    elif alert_mode == 'min':
      return np.min, 'mi'
    elif alert_mode == 'max':
      return np.max, 'Mx'
    raise ValueError("Evaluation function unknown '{}'".format(alert_mode))

  def _reset_queue(self, hard=False):    
    new_queue = deque(maxlen=self._queue_size)
    new_time = deque(maxlen=self._queue_size)
    if hasattr(self, 'queue') and len(self.queue) > 0 and not hard:
      new_queue.append(self.queue[-1])
      new_time.append(self._times_queue[-1])
    self.queue = new_queue
    self._times_queue = new_time

    new_values_queue = deque(maxlen=self._queue_size)
    if hasattr(self, '_raw_values_queue') and len(self._raw_values_queue) > 0 and not hard:
      new_values_queue.append(self._raw_values_queue[-1])
    self._raw_values_queue = new_values_queue

    self._queue_time = time() # FIXME: this can be set to None
    self._first_change = True
    self.in_confirmation = False
    self._confirmation_wait = 0
    self._last_changing_state = self._changing_state
    self._changing_state = None
    return
  
  def _increase_queue(self):
    curr_len = self.queue.maxlen
    new_queue = deque(maxlen=curr_len + 1)
    new_queue.extend(self.queue)
    self.queue = new_queue
    
    new_values_queue = deque(maxlen=curr_len + 1)
    new_values_queue.extend(self._raw_values_queue)
    self._raw_values_queue = new_values_queue    

    new_times_queue = deque(maxlen=curr_len + 1)
    new_times_queue.extend(self._times_queue)
    self._times_queue = new_times_queue    
    return
  
  
  def _queue_full(self):
    return len(self.queue) >= self._queue_size
  
  
  def _confirmation_time_passed(self, return_time=False):
    # TODO: review if there we should use self._queue_time or self.get_queue_time()
    _time = time() - self._queue_time
    self._confirmation_wait = _time
    self._last_confirmation_wait = _time
    if self._state:
      # if alert then (bigger) wait before lowering
      _passed = _time >= self._lower_wait_time      
    else:
      # wait before raising alert
      _passed = _time >= self._raise_wait_time
    if return_time:
      return _passed, _time
    return _passed
  
  
  def reset(self, default=False, hard_reset=False):
    self._state = default
    self._last_state = self._state
    self._reset_queue(hard=hard_reset)
    return
  
  
  def hard_reset(self, default=False):
    self.reset(
      default=default,
      hard_reset=True,
    )
    return
  
  
  def get_queue_time(self):
    if self._eval_times is not None and len(self._eval_times) > 1:
      qt = self._eval_times[-1] - self._eval_times[0]
      return qt
    return None
  
  def get_last_raw_value(self):
    if len(self._raw_values_queue) > 0:
      return self._raw_values_queue[-1]
    return None
  
  def get_last_eval_queue(self):
    return self._eval_queue if self._eval_queue is not None else []

  def get_last_eval_queue_raw(self):
    return self._eval_queue_raw if self._eval_queue_raw is not None else []
      
  def get_last_confirmation_time(self):
    return round(self._last_confirmation_wait,2)
 
  
  def add_observation(self, value, reduce_threshold=None):
    if reduce_threshold is None:
      reduce_threshold = self._reduce_threshold
    if value is None:
      return
    # add observation    
    self._raw_values_queue.append(value)
    self._times_queue.append(time())
    if self._reduce_value:
      self.queue.append(int(value >= reduce_threshold))
    else:
      self.queue.append(value)
    self._eval_queue = self.queue.copy()
    self._eval_times = self._times_queue.copy()
    self._eval_queue_raw = self._raw_values_queue.copy()
        
    # check if enough data is accumulated
    if self._queue_full():
      # compute current status
      alert_value = self._lower_alert_value if self._state else self._raise_alert_value
      eval_func = self._eval_func_lower if self._state else self._eval_func
      current_state = bool(eval_func(self.queue) >= alert_value)
        
      # check if this was first change after reset
      if current_state != self._state and self._first_change:
        self._first_change = False
        # now we can wait until enough time has passed
        self._queue_time = time()
      #endif
    
      last_state = self._state
      # check if new proposed state is different than last proposed state 
      # and enough time has passed from last change passed (that forces a extension
      # of the queue if required)
      if (current_state != self._state):
        self._changing_state = current_state
        confirmed, waited = self._confirmation_time_passed(True)
        self.in_confirmation = not confirmed
        if confirmed:
          self._state = current_state
          self._reset_queue()
          self._change_time = self._times_queue[-1]
          if self._state:
            self._last_raise_timestamp = self._change_time
          else:
            self._last_lower_timestamp = self._change_time
        #endif

      # we do this check after each iteration because we need to extend the queue
      # no matter what the current_state is. If we extend the queue only when the
      # current state == last state then we discard observations that enforce the change
      # after it has taken place (see the spike neg test)
      if self.in_confirmation:
        if self._confirmation_time_passed():
          self._reset_queue()
        else:
          # we extend the buffer in order to evaluate all observations from the
          # first change until now. If we increase the queue only if the state 
          # is changed then we can have spikes that trigger false alerts 
          # Run the spike test with the previous version
          self._increase_queue()
      #endif
      self._last_state = last_state
    else:
      self._last_state = self._state
    return
  
  def status_changed(self):
    return self._last_state != self._state
  
  def is_alert(self):
    return self._state
  
  def is_new_alert(self):
    return self.is_alert() and self.status_changed()

  def is_new_raise(self):
    return self.is_new_alert()

  def is_new_lower(self):
    return not self.is_alert() and self.status_changed()
  
  def get_time_from_change(self):
    return time() - self._change_time if self._change_time is not None else -1
  
  
  def get_last_alert_duration(self):
    """
    This function returns the time spent during the last alert. Function workd only if the alert is not active

    Returns
    -------
      float

    """
    result = None
    last_raise = self._last_raise_timestamp
    last_lower = self._last_lower_timestamp
    has_raise = last_raise is not None
    has_lower = last_lower is not None
    if has_raise and has_lower and last_lower > last_raise:
      result = round(last_lower - last_raise, 2)
    return result
  
  

  def get_setup_values(self):
    dct = {
      'queue_size': self._queue_size,
      'raise_wait_time': self._raise_wait_time,
      'lower_wait_time': self._lower_wait_time,
      'raise_alert_value': self._raise_alert_value,
      'lower_alert_value': self._lower_alert_value,
      'alert_mode': self._alert_mode,
      'alert_mode_lower': self._alert_mode_lower,
      'reduce_value': self._reduce_value,
      'reduce_threshold': self._reduce_threshold
    }
    return dct

  def __repr__(self):
    _q = [round(x, 3) for x in self.get_last_eval_queue()]
    _q_raw = [round(x, 3) for x in self.get_last_eval_queue_raw()]
        
    if self._reduce_value:
      s_q = "["+ ", ".join(["{}:{}".format(v1, v2) for v1,v2 in zip(_q, _q_raw)])
      s_q = s_q + "]"
    else:
      if len(_q) > MAX_QUEUE_SHOW:
        s_q = "<{}>[...,".format(len(_q))
        s_q = s_q + ",".join([str(xx) for xx in _q[-7:]])
        s_q = s_q + "]"
      else:
        s_q = str(_q)
    
    _ct = 'NA'
    if self.status_changed():
      _ct = "{}/{}".format(
        self.get_last_confirmation_time(), 
        self._raise_wait_time if self._state else self._lower_wait_time,
        )
    elif self.in_confirmation:
      _ct = "{}/{}".format(
        self.get_last_confirmation_time(), 
        self._raise_wait_time if not self._state else self._lower_wait_time,
        )
    else:
      # no confirmation, no status changed ... but we need to test if we just droped
      # a false status changed
      if len(_q) > self._queue_size:
        _ct = "{}/{}".format(
          self.get_last_confirmation_time(), 
          self._raise_wait_time if self._last_changing_state else self._lower_wait_time,
          )

    params_idx = self.is_alert() if self.status_changed() else not self.is_alert()
    alert_value = [self._lower_alert_value, self._raise_alert_value][params_idx]
    vs_oper = ['<', '>='][params_idx]
    eval_method = [self._eval_method_lower, self._eval_method][params_idx]
    eval_func = [self._eval_func_lower, self._eval_func][params_idx]
    # endif long code vs short code
    
    queue_time = self.get_queue_time()
    last_change = self.get_time_from_change()

    _s = "A={}, N={}, CT={}, E={}{}={:.2f} {}vs {}{:.2f} {}{}".format(
      int(self.is_alert()),
      int(self.status_changed()),
      _ct,
      eval_method,
      s_q,
      eval_func(_q) if len(_q) > 0 else 0,
      "" if queue_time is None else "(in {:.1f}s) ".format(queue_time),
      vs_oper,
      alert_value,
      "" if last_change == -1 else "LstCh:{:.1f}s ".format(last_change),
      "({} v{})".format(self.name, self.version) if self._show_version else '',
      )
    return _s
    
    
  
if __name__ == '__main__':
  from time import sleep, strftime, localtime
  
  def time_to_str(self, t=None, fmt='%Y-%m-%d %H:%M:%S'):
    if t is None:
      t = time()
    return strftime(fmt, localtime(t))
  LOOP_RESOLUTION = 4
  LOOP_TIME = 1 / LOOP_RESOLUTION
  DATA_TRAIN = [0,4,3,1,0,0,0,5,9,8,7,1,1,1,1,0,0,0,0,0,0]
  DATA_TRAIN_PRC = [
    0.31,0.6,0.5,0.5,0.2,0.3,0.4,0.515,
    0.9,0.8,0.71,0.51,0.61,0.71,0.49,
    0.35,0.1,0.1,0.1,0.1,0.1,0.1,0.1,0.1,0.1,0.1,0.1,0.2,0.4,
    0.5,0.9,0.8,0.71,0.51,0.61,0.71,0.51,
    ]
  DATA_TRAIN_SPIKE = [
    100, 100, 100, # trigger a spike
    *([0] * (LOOP_RESOLUTION - 1)),
    *([0] * 5 * LOOP_RESOLUTION),
    0,
    *([70] * 4),
    45,
    70,
    *([0] * (LOOP_RESOLUTION - 1)),
    ]
  
  DATA_TRAIN_SPIKE_NEG = [
    *([100] * LOOP_RESOLUTION),
    *([100] * LOOP_RESOLUTION),
    *([0] * LOOP_RESOLUTION),
    *([100] * LOOP_RESOLUTION),
    *([100] * LOOP_RESOLUTION),
    *([100] * LOOP_RESOLUTION),
    *([0] * LOOP_RESOLUTION),
    *([0] * LOOP_RESOLUTION),
    *([100] * LOOP_RESOLUTION),
  ]

  TEST_COUNTS = False
  TEST_BOOLS = False
  TEST_PERCENT = False
  TEST_SPIKE = False
  TEST_SPIKE_NEG = True
  
  if TEST_PERCENT:
    print("Testing percent alerts ...")
    asm = AlertHelper(
      name='TEST_PLUGIN_01',
      values_count=1,
      raise_confirmation_time=2,
      lower_confirmation_time=2,
      reduce_value=False,
      reduce_threshold=0.5,
      alert_mode='mean',
      raise_alert_value=0.5,
      lower_alert_value=0.4,
      show_version=False,
      )
    
    for i, value in enumerate(DATA_TRAIN_PRC):
      t = time_to_str(time())
      asm.add_observation(value)
      if asm.is_new_alert():
        msg = "[New alert]"
      elif asm.is_new_lower():
        msg = "[Lwr alert]"
      elif not asm.is_alert():
        msg = "[No change]"
      print(f"{t} {msg} {asm.get_last_alert_duration()} {asm}")
      sleep(LOOP_TIME)
  
  if TEST_SPIKE:
    # in this test, there is a spike in the data train 
    # that triggers a confirmation time check and another spike
    # when the confirmation is almost over
    
    # this test checks that spikes do not trigger false alerts
    # the alert should not be raised after the confirmation time
    print("Testing spike alerts ...")
    asm = AlertHelper(
      name='TEST_PLUGIN_01',
      values_count=5,
      raise_confirmation_time=7,
      lower_confirmation_time=30,
      reduce_value=False,
      reduce_threshold=0.5,
      alert_mode='mean',
      raise_alert_value=60,
      lower_alert_value=30,
      show_version=False,
      )
    
    for i, value in enumerate(DATA_TRAIN_SPIKE):
      t = time_to_str(time())
      asm.add_observation(value)
      if asm.is_new_alert():
        msg = "[New alert]"
      elif asm.is_new_lower():
        msg = "[Lwr alert]"
      elif not asm.is_alert():
        msg = "[No change]"
      print(f"{t} {msg} {asm.get_last_alert_duration()} {asm}")
      sleep(LOOP_TIME)

  if TEST_SPIKE_NEG:
    # in this test, there is a spike in the data train 
    # that temporarily lowers the alert status
    
    # this test checks that once a confirmation time starts, 
    # all observations are added to the queue, regardless of 
    # the current state of the alert, and the alert is computed 
    # on the whole sequence
    # the alert should be raised after the confirmation time
    print("Testing negative spike alerts ...")
    asm = AlertHelper(
      name='TEST_PLUGIN_01',
      values_count=5,
      raise_confirmation_time=7,
      lower_confirmation_time=30,
      reduce_value=False,
      reduce_threshold=0.5,
      alert_mode='mean',
      raise_alert_value=60,
      lower_alert_value=30,
      show_version=False,
      )
    
    for i, value in enumerate(DATA_TRAIN_SPIKE_NEG):
      t = time_to_str(time())
      asm.add_observation(value)
      if asm.is_new_alert():
        msg = "[New alert]"
      elif asm.is_new_lower():
        msg = "[Lwr alert]"
      elif not asm.is_alert():
        msg = "[No change]"
      print(f"{t} {msg} {asm.get_last_alert_duration()} {asm}")
      sleep(LOOP_TIME)



import gc
import os
from pathlib import Path

import torch as th
import numpy as np
import json
import tensorrt as trt

from typing import Any, Mapping, Tuple, Sequence, Union

from naeural_core import constants as ct
from naeural_core.serving.base.backends.model_backend_wrapper import ModelBackendWrapper

def get_torch_dtype(np_dtype : np.dtype) -> th.dtype:
  """
  Get a torch dtype from a numpy dtype.
  Parameters:
    - np_dtype a numpy dtype
  Returns:
    - the corresponding torch dype.
  """
  numpy_to_torch_dtype_dict = {
      bool          : th.bool,
      np.bool_      : th.bool,
      np.uint8      : th.uint8,
      np.int8       : th.int8,
      np.int16      : th.int16,
      np.int32      : th.int32,
      np.int64      : th.int64,
      np.float16    : th.float16,
      np.float32    : th.float32,
      np.float64    : th.float64,
      np.complex64  : th.complex64,
      np.complex128 : th.complex128
  }
  if np_dtype not in numpy_to_torch_dtype_dict.keys():
    raise ValueError("No known corresponding torch dtype for {}".format(np_dtype))
  return numpy_to_torch_dtype_dict[np_dtype]


def check_trt_version(
    trt_version : str,
    cuda_version : str,
    model_device_name : str,
    current_device : th.device):
  """
  Checks the provided version strings and device against the current environment.
  Parameters:
    - trt_version : str, TensorRT version
    - cuda_version : str, CUDA version string
    - model_device_name : str, GPU name
    - current_device : th.device, the checked torch device
  Returns:
    bool, True if this is a compatible configuration,, False otherwise.
  """
  if trt_version != trt.__version__ or cuda_version != th.version.cuda:
    return False
  if model_device_name != th.cuda.get_device_name(current_device):
    return False
  return True

class TRTServingLogger(trt.ILogger):
  """
  A TensorRT-compatible wrapper around the EE logger so that TensorRT
  can provide logs.
  """
  def __init__(self, logger):
    trt.ILogger.__init__(self)
    self._logger = logger
    return

  def log(self, severity : trt.ILogger.Severity, msg : str):
    if severity == trt.ILogger.VERBOSE:
      # Skip verbose messages.
      return
    color = ct.COLORS.SERVING
    self._logger.P('[TRT] ' + msg, color=color)
    return

  def get_base_logger(self):
    return self._logger


class TensorRTModel(ModelBackendWrapper):
  """
  Wrapper around TensorRT for inference.
  For further information see the TensorRT documentation at
  https://docs.nvidia.com/deeplearning/tensorrt/archives/tensorrt-1001/api/python_api/
  """

  # Metadata keys
  METADATA_MAX_BATCH = 'max_batch_size'
  METADATA_INPUTS_KEY = 'inputs'
  METADATA_OUTPUTS_KEY = 'outputs'
  METADATA_ONNX_MD5 = 'onnxmd5' # ONNX md5
  METADATA_ENGINE_MD5 = 'enginemd5' # ONNX md5

  # TensorRT .engine compatibility metadata.
  # We need to check these values against the current environment
  # to ensure that the .engine file can run.
  METADATA_CUDA_VERSION_KEY = 'cudaver' # CUDA version identifier metadata
  METADATA_TRT_VERSION_KEY = 'trtver' # TensorRT identifier metadata
  METADATA_CUDA_DEVICE_KEY = 'cudadev' # GPU device name
  METADATA_PRECISION_KEY = 'precision'

  ONNX_METADATA_KEY = 'onnx_metadata'


  def _clear_internal_state(self):
    self._context = None # TensorRT context
    self._model = None # TensorRT model
    self._metadata = None # Model metadata (JSON)
    self._trt_to_th_idx = None # Permutation from TensorRT to torch order
    self._output_args = None # Tuple of tensor outputs (in torch order)
    self._max_batch_size = None # The maximum batch size the model was built with
    self._current_batch_size = None # The last batch size used for inference
    self._batched_input_idx = None # The index of a batched input
    self._batch_runtime_info = None # Runtime batch-size dependent information
    self._device = None # The device for the model
    return


  def __init__(self, log):
    self._clear_internal_state()
    self._log = TRTServingLogger(log)
    # Make sure the TRT plugins are initialized.
    trt.init_libnvinfer_plugins(None, "")
    return


  def get_device(self):
    return self._device


  def _allocate_outputs(self, batch_size : int) -> Tuple[th.tensor, ...]:
    """
    Allocate output tensors for the current model batch size.
    Parameters:
      batch_size : int, the batch size to use when allocating the output tensors.

    Returns:
      A tuple of torch tensors for the current batch size.
    """
    _, output_types = self._batch_runtime_info[batch_size]
    outputs = []
    for out_shape, dtype in output_types:
      outputs.append(th.empty(
        size=out_shape,
        dtype=dtype,
        device=self._device,
        requires_grad=False
      ).contiguous())
    return tuple(outputs)

  def _set_batch_size(self, batch_size : int):
    """
    Set the current dynamic shape to batch_size.
    We need to call this every time the batch size changes.
    Parameters:
      batch_size: integer, the current batch size
    Returns: None
    """
    self._current_batch_size = batch_size
    shape_info, _ = self._batch_runtime_info[batch_size]
    for idx, name, shape in shape_info:
      # Update all batch-dependent tensor with their new shape.
      if not self._context.set_input_shape(name, shape):
        raise ValueError("Unable to set dynamic shape")

    return


  def _deserialize_model(
    self,
    model_path : str,
    device : th.device):
    """
    Initialize the model and metadata from the TensorRT model serialized
    on disk at the path model_path.
    Parameters:
      model_path : string, the path to the TensorRT model on the disk.
    Returns: None
    """
    with open(model_path, 'rb') as f, trt.Runtime(self._log) as runtime:
      # Set the CUDA device to device so that the model will be loaded to it.
      # It's not recommended to call this function in general, but according to
      # the docs this is the only way to load to a certain device. Only do this
      # if we have more than one device since otherwise we can run into issues
      # with the device index being set to None.
      # See https://docs.nvidia.com/deeplearning/tensorrt/developer-guide/index.html#faq
      if th.cuda.device_count() > 1:
        th.cuda.set_device(device)
      # Now read the actual TensorRT model.
      self._model = runtime.deserialize_cuda_engine(f.read())  # read engine
    #endwith read model from disk
    return


  def _get_torch_args_mapping(self):
    """
    Get the mapping of the input/output arguments to TensorRT order.
    """
    th_arg_positions = {}
    for i, name in enumerate(self._input_names + self._output_names):
      th_arg_positions[name] = i

    num_bindings = self._model.num_io_tensors
    return [
      th_arg_positions[self._model.get_tensor_name(i)]
      for i in range(num_bindings)
    ]


  def _get_batch_dependent_arg_idx(self) -> int:
    """
    Get the index of an input argument whose size is batch-dependent.
    That is the shape of that argument should be batch_size x ...
    """
    for i, name in enumerate(self._input_names):
      shape = self._context.get_tensor_shape(name)
      if shape[0] == -1:
        return i
    # No batching?
    return None


  def _get_batch_runtime_state(self, batch_size : int):
    """
    Get the current runtime state dependent on the batch size

    """
    # Produce a list of bindings indices and shapes
    # for the bindings that need to be set once the
    # batch size changes
    # Here will be cached only the dynamic batches, hence
    # in case of a dynamic shape the engine may change at runtime
    # (e.g. at the warmup stage when the batch size may vary).
    dynamic_shapes = []
    for i, name in enumerate(self._input_names):
      shape = self._context.get_tensor_shape(name)
      if shape[0] != -1:
        continue
      shape = trt.Dims([batch_size] + list(shape[1:]))
      dynamic_shapes.append((i, name, shape))
      #endfor all batch sizes
    #endfor all bindings

    # Produce a list of output tensors shapes for this batch size.
    # There will be cached info for what the allocated memory should be.
    # Thus, we cache the fixed batch sizes too.
    batch_output_types = []
    for name in self._output_names:
      shape = self._context.get_tensor_shape(name)
      shape = tuple([batch_size if size == -1 else size for size in list(shape)])
      np_dtype = trt.nptype(self._model.get_tensor_dtype(name))
      dtype=get_torch_dtype(np_dtype)
      batch_output_types.append((shape, dtype))
    #endfor all output tensors

    return dynamic_shapes, batch_output_types

  def check_engine_file(
    self,
    device : th.device,
    onnx_path,
    engine_metadata_path,
    engine_path,
    half : bool,
    max_batch_size : int
  ) -> None:
    """
    Parameters:
     - device - torch device on which the model will be loaded.
     - onnx_path - path to the onnx model
     - engine_metadata_path - path to the local engine metadata file
         produced by create_from_onnx
     - engine_path - path to the TensorRT engine file
     - max_batch_size - int, maximum batch size to use. If this value
        is different from the one stored in the engine metadata file
        the check will fail and a ValueError will be raised.

    Returns: None

    Raises:
     - ValueError if this configuration cannot be loaded
    """
    if not os.path.isfile(str(engine_metadata_path)):
      raise ValueError("No metadata file")

    try:
      with open(str(engine_metadata_path), 'rt') as f:
        metadata = json.load(f)
    except json.JSONDecodeError as _:
      raise ValueError("Invalid engine config JSON")

    if not self.METADATA_INPUTS_KEY in metadata.keys():
      raise ValueError("Missing input names metadata")

    if not self.METADATA_OUTPUTS_KEY in metadata.keys():
      raise ValueError("Missing input names metadata")

    if not self.METADATA_ONNX_MD5 in metadata.keys():
      raise ValueError("Missing onnx md5 metadata")

    if not self.METADATA_MAX_BATCH in metadata.keys():
      raise ValueError("Missing max_batch_size metadata")

    trt_version = metadata[self.METADATA_TRT_VERSION_KEY]
    cuda_version = metadata[self.METADATA_CUDA_VERSION_KEY]
    device_name = metadata[self.METADATA_CUDA_DEVICE_KEY]

    if max_batch_size != metadata[self.METADATA_MAX_BATCH]:
      raise ValueError("Batch size mismatch")

    if not check_trt_version(trt_version, cuda_version, device_name, device):
      # Bail out here, we have a version mismatch so an engine
      # rebuild is required.
      raise ValueError("Version mismatch")

    import hashlib
    if not os.path.isfile(str(onnx_path)):
      raise ValueError("No ONNX file")
    if not os.path.isfile(str(engine_path)):
      raise ValueError("No .engine file")

    with open(str(onnx_path), 'rb') as f:
      onnx_md5 = hashlib.md5(f.read()).hexdigest()
    with open(str(engine_path), 'rb') as f:
      engine_md5 = hashlib.md5(f.read()).hexdigest()

    if metadata[self.METADATA_ONNX_MD5] != onnx_md5:
      raise ValueError("ONNX MD5 mismatch")
    if metadata[self.METADATA_ENGINE_MD5] != engine_md5:
      raise ValueError("Engine MD5 mismatch")

    if (metadata[self.METADATA_PRECISION_KEY] == "fp16") != half:
        raise ValueError("Precision mismatch")

    return

  @staticmethod
  def _get_engine_and_config_path(
    onnx_path : str,
    half : bool,
    max_batch_size : int
  ) -> Tuple[str, str]:
    """
    Get the .engine and config file path given an ONNX model.

    Parameters:
      - onnx_path, str, path to the .onnx model
      - half : bool, True if engine is built for fp16
      - max_batch_size : int, maximum batch size supported in inference

    Returns:
      Tuple(str, str) containing the .engine and configuration file
      paths.
    """
    onnx_file_basename = os.path.basename(onnx_path)
    onnx_file_path = Path(onnx_path)
    if half:
      precision = 'fp16'
    else:
      precision = 'fp32'

    config_rel_path = os.path.join(precision, 'bs' + str(max_batch_size))

    onnx_conf_path = Path(os.path.join(
      os.path.dirname(str(onnx_file_path)),
      config_rel_path,
      onnx_file_basename
    ))
    engine_path = onnx_conf_path.with_suffix(".engine")
    engine_metadata_path = onnx_conf_path.with_suffix(".engine.json")
    return str(engine_path), str(engine_metadata_path)

  def load_or_rebuild_model(
    self,
    onnx_path : str,
    fp16 : bool,
    max_batch_size : int,
    device : th.device
  ) -> None:
    """
    Loads a TensorRT engine from engine_path if possible.
    If not possible (e.g. due to the file not existing, cuda/tensor-rt
    version mismatches, etc) we rebuild the TensorRT engine from the
    ONNX model located at engine_path

    Parameters:
      onnx_path - string, path of the ONNX model on disk
      fp16: bool, True if the model should be casted to float16
      max_batch_size - the maximum batch size to use in inference
      device - torch device to use.

    Returns:
      None
    """

    onnx_file_path = Path(onnx_path)
    engine_path, engine_metadata_path = self._get_engine_and_config_path(
      onnx_path,
      fp16,
      max_batch_size
    )

    try:
      self.check_engine_file(
        device,
        onnx_file_path,
        engine_metadata_path,
        engine_path,
        fp16,
        max_batch_size
      )

      # Try to load the model from the disk.
      self.load_model(engine_path, engine_metadata_path, device)
      # Model was loaded so there is nothing left to do.
      return
    except ValueError as ve:
      # If check_engine_file passes we except to be able to
      # load the model, so just catch ValueErrors from it,
      self._log.log(severity=trt.ILogger.INFO, msg=str(ve))

    # Clear any kind of left-over internal state from the failed load_model.
    self._clear_internal_state()

    # Reconstruct the TensorRT engine file.
    self._log.log(
      severity=trt.ILogger.INFO,
      msg="Failed to load model, trying to rebuild"
    )
    self._log._logger.P(
      "Building TensorRT engine file, this can take up to an hour",
      color='r'
    )
    TensorRTModel.create_from_onnx(
      onnx_path=onnx_path,
      half=fp16,
      max_batch_size=max_batch_size,
      device=device,
      log=self._log.get_base_logger()
    )
    # There is no need to check for a ValueError here since we
    # expect create_from_onnx to give us something that will pass
    # this check. If this is not the case a crash is appropriate.
    self.check_engine_file(
      device,
      onnx_file_path,
      engine_metadata_path,
      engine_path,
      fp16,
      max_batch_size
    )
    self.load_model(engine_path, engine_metadata_path, device)
    return

  def get_input_dtype(self, idx):
    name = self._input_names[idx]
    np_dtype = trt.nptype(self._model.get_tensor_dtype(name))
    th_dtype = get_torch_dtype(np_dtype)
    return th_dtype

  def load_model(
    self,
    model_path : str,
    metadata_path : str,
    device : th.device
  ) -> None:
    """
    Initialize from a path on the disk. The initialization may raise
    exceptions if the cuda or TensorRT versions mismatch.

    Parameters:
      model_path - path to tensor-rt model on disk
      device - torch device to use.

    Returns: None
    """

    if self._context is not None:
      raise RuntimeError("Model already loaded")

    with open(metadata_path, 'rt') as f:
      self._metadata = json.load(f)

    self._deserialize_model(model_path, device)

    self._context = self._model.create_execution_context()
    self._input_names = self._metadata[TensorRTModel.METADATA_INPUTS_KEY]
    self._output_names = self._metadata[TensorRTModel.METADATA_OUTPUTS_KEY]
    self._max_batch_size = self._metadata[TensorRTModel.METADATA_MAX_BATCH]
    self._trt_to_th_idx = self._get_torch_args_mapping()
    self._device = device

    # Initialize index of the first batch dependent input.
    self._batched_input_idx = self._get_batch_dependent_arg_idx()

    # Pre-construct shapes and outputs for all possible batch sizes.
    # This avoids doing work at inference time.
    self._batch_runtime_info = []
    for batch_size in range(self._max_batch_size + 1):
      self._batch_runtime_info.append(self._get_batch_runtime_state(batch_size))

    # Finally set the current batch size to the maximum batch size.
    self._set_batch_size(self._max_batch_size)
    return


  def forward(self, *args : th.tensor) -> Tuple[th.tensor, ...]:
    """
    Run inference. All model inputs and outputs are torch tensors.

    Parameters:
      All parameters are input tensors in the order specified when the model
      was exported from torch. All parameters are unnamed.

    Returns:
      A tuple of output tensors, in the order of outputs of the
      original torch model. Outputs are not captured by the model
      and are owned by the caller. In case of a tuple with a single
      output, the output is returned directly.
    """
    batch_size = self._current_batch_size
    # If the batch size was changed from the last inference
    # we need to let TensorRT know.
    if self._batched_input_idx is not None:
      batch_size = args[self._batched_input_idx].shape[0]
      if batch_size != self._current_batch_size:
        self._set_batch_size(batch_size)
    #endif set batch size if changed

    # Allocate output tensors.
    current_output_tensors = self._allocate_outputs(batch_size)

    # Add the implicit (output) arguments to the list of input arguments.
    # Make the inputs contiguous if they are not since TensorRT requires this.
    # Output tensors have already been allocated in a contiguous fashion on
    # creation.
    args_list = [x.contiguous() for x in args] + list(current_output_tensors)

    # Convert the args list into the actual tensor-rt parameter list.
    # This accounts for the difference in argument ordering between torch
    # and TensorRT.
    parameter_addrs = [
      int(args_list[self._trt_to_th_idx[i]].data_ptr())
      for i in range(self._model.num_io_tensors)
    ]

    if not self._context.execute_v2(parameter_addrs):
      raise Exception("Failed to run TensorRT kernel")

    # Return pre-computed output tensors in original ordering.
    if len(current_output_tensors) == 1:
      return current_output_tensors[0]
    return current_output_tensors


  def __call__(self, *args : th.tensor) -> Tuple[th.tensor, ...]:
    """
    Override for the call operator, just goes to forward.
    """
    return self.forward(*args)


  @staticmethod
  def create_from_onnx(
    onnx_path : str,
    half : bool,
    max_batch_size : int,
    device : th.device,
    log,
    workspace_size : int = 4,
    flags : Sequence[trt.NetworkDefinitionCreationFlag] =
      [trt.NetworkDefinitionCreationFlag.EXPLICIT_BATCH],
    builder_flags : Sequence[trt.BuilderFlag] = [],
  ) -> None:
    """
    Export an ONNX model to a tensor-rt model on disk.
    The engine model and the metadata file will be written with
    extensions '.engine' and '.engine.json' respectively.
    Input and output names must be valid python identifiers.

    Parameters:
      onnx_path - string, path on disk to the input onnx model
      half - bool, exports the model to fp16 if True
      max_batch_size - int, maximum batch size for the produced
        engine.
      log - the EE logger object
      workspace_size - size of the Tensor-RT workspace (in GB)
        Defaults to 4.
      flags - a list of TensorRT NetworkDefinitionCreationFlag flags.
        Defaults to [NetworkDefinitionCreationFlag.EXPLICIT_BATCH]
      builder_flags - a list of TensorRT BuilderFlag. If half is true
        the FP16 flag is implicitly added. Defaults to [].

    Returns: None
    """

    import onnx
    import hashlib
    from google.protobuf.json_format import MessageToDict

    trt_logger = TRTServingLogger(log)

    path, config_path = TensorRTModel._get_engine_and_config_path(
      onnx_path,
      half,
      max_batch_size
    )
    # Create folders if they don't exist.
    os.makedirs(os.path.dirname(path), exist_ok=True)

    # Load the ONNX model from disk just to get the input/output names
    # in the correct order.
    model_onnx = onnx.load(onnx_path)

    with open(onnx_path,'rb') as f:
      onnx_md5 = hashlib.md5(f.read()).hexdigest()

    # Gather input and output names from the onnx model.
    # We need to save these in order in the TensorRT engine
    # metadata.
    def get_onnx_names(onnx_nodes_list):
      ret = []
      for name in onnx_nodes_list:
        message_dict = MessageToDict(name)
        ret.append(message_dict['name'])
      return ret

    output_names = get_onnx_names(model_onnx.graph.output)
    input_names = get_onnx_names(model_onnx.graph.input)

    # Extract the onnx metadata from the onnx model.
    onnx_metadata_dict = MessageToDict(model_onnx, preserving_proto_field_name = True)['metadata_props']
    for meta in onnx_metadata_dict:
      if meta['key'] == TensorRTModel.ONNX_METADATA_KEY:
        onnx_metadata = json.loads(meta['value'])
        break

    model_precision = onnx_metadata.get('precision')
    # TODO: change this to a maybe_change_precision function that will be situated in a utils file
    if model_precision is not None:
      if (model_precision.lower() == "fp16") != half:
        # We're raising a runtime error specifically so that this doesn't
        # get caught by load_or_rebuild_model
        if half and model_precision.lower() == "fp32":
          #from onnxconverter_common import float16
          from onnxruntime.transformers import float16
          model_fp16 = float16.convert_float_to_float16(model_onnx)
          fp16path = str(Path(onnx_path).with_suffix('.tofp16.onnx'))
          onnx.save(model_fp16, fp16path)
          onnx_path = fp16path
        else:
          raise RuntimeError('Incompatible precision in ONNX model')
    #endif check onnx precision

    metadata = {
      TensorRTModel.METADATA_ONNX_MD5 : onnx_md5,
      TensorRTModel.METADATA_INPUTS_KEY : input_names,
      TensorRTModel.METADATA_OUTPUTS_KEY : output_names,
      TensorRTModel.METADATA_MAX_BATCH : max_batch_size,
      TensorRTModel.METADATA_TRT_VERSION_KEY : trt.__version__,
      TensorRTModel.METADATA_CUDA_VERSION_KEY : th.version.cuda,
      TensorRTModel.METADATA_CUDA_DEVICE_KEY : th.cuda.get_device_name(device)
    }
    # We're done with the ONNX model so nuke it now.
    del model_onnx
    model_onnx = None

    # Set up the tensor-rt framework boilerplate.
    builder = trt.Builder(trt_logger)
    config = builder.create_builder_config()

    # Set the workspace scratch memory size. Note that this is only
    # for one layer and the engine can use much more overall in an
    # execution.
    config.set_memory_pool_limit(
      trt.MemoryPoolType.WORKSPACE,
      pool_size=workspace_size * (1 << 30)
    )

    # Build the flag mask for the network.
    flag_mask = 0
    for flag in flags:
      flag_mask |= (1 << int(flag))

    # Load model through tensor-rt's ONNX parser.
    network = builder.create_network(flag_mask)
    parser = trt.OnnxParser(network, trt_logger)
    if not parser.parse_from_file(onnx_path):
      raise RuntimeError(f'Failed to load ONNX file: {onnx_path}')
    if half:
      config.set_flag(trt.BuilderFlag.FP16)
    for config_flag in builder_flags:
      config.set_flag(config_flag)

    # Add the optimization profile.
    profile = builder.create_optimization_profile()
    inputs = [network.get_input(i) for i in range(network.num_inputs)]
    for input in inputs:
      if input.shape[0] != -1:
        # Tensor doesn't have a dynamic batch dimension, ignore.
        continue
      # Add an optimization profile for this input tensor.
      # Optimize for the max batch size case.
      min_shape = [1] + list(input.shape[1:])
      opt_shape = [max_batch_size] + list(input.shape[1:])
      max_shape = [max_batch_size] + list(input.shape[1:])
      profile.set_shape(
        input.name,
        min=min_shape,
        opt=opt_shape,
        max=max_shape
      )
    #endfor all inputs
    config.add_optimization_profile(profile)

    # Empty cache before building the engine.
    gc.collect()
    th.cuda.empty_cache()

    # Write built engine to disk.
    timer_build_id = "trt_engine_build"
    log.start_timer(timer_build_id)
    try:
      if not hasattr(builder, 'build_engine'):
        # This is the TensorRT 10.0 API case where we don't have a build_engine
        with builder.build_serialized_network(network, config) as engine, open(path, "wb") as t:
          t.write(engine)
      else:
        # This is the base case for TensorRT 8.6.1
        with builder.build_engine(network, config) as engine, open(path, 'wb') as t:
          # Write the serialized TensorRT engine.
          t.write(engine.serialize())
      log.stop_timer(timer_build_id)
      log.P("TensorRT model build took {}".format(log.get_timer_mean(timer_build_id)))
    except Exception as e:
      log.stop_timer(timer_build_id)
      raise RuntimeError(f"Failed to build TensorRT engine: {e}")

    with open(path,'rb') as f:
      engine_md5 = hashlib.md5(f.read()).hexdigest()

    metadata[TensorRTModel.METADATA_ENGINE_MD5] = engine_md5
    # Also record the ONNX metadata
    metadata[TensorRTModel.ONNX_METADATA_KEY] = onnx_metadata
    # Record the actual precision of the model
    metadata[TensorRTModel.METADATA_PRECISION_KEY] = 'fp16' if half else 'fp32'

    with open(config_path, 'wt') as f:
      json.dump(metadata, f)
    #endif on status write metadata file
    return



def solution(S, K):
  DAYS = ['Mon', 'Tue', 'Wed', 'Thu', 'Fri', 'Sat','Sun']
  assert S in DAYS  
  idx = DAYS.index(S)
  i = (idx + K) % len(DAYS)
  return DAYS[i]



print(solution('Wed', 2))
print(solution('Sat', 23))



def solution1(S):
  from itertools import groupby
  lst_case = [0 for _ in range(len(S))]
  for j,c in enumerate(S):
    if c.isupper():
      for i,_c in enumerate(S):
        if _c == c.lower():
          lst_case[i] = 2
          lst_case[j] = 2
    else:
      for i,_c in enumerate(S):
        if _c == c.upper():
          lst_case[i] = 2
          lst_case[j] = 2
  l = [list(group) for item, group in groupby(lst_case) if item == 2]
  if len(l) > 0:
    lst = max(l, key=len)
    if len(lst) > 1:
      return len(lst)
    else:
      return -1
  else:
    return -1
  
def solution2(s):
    """
    Given a string `s` of up to 200 characters, returns the longest balanced substring.
    A balanced substring is defined as a sequence of letters where each letter appears both in lower and upper case.

    Args:
        s (str): The input string.

    Returns:
        str: The longest balanced substring.
    """
    # Initialize a dictionary to keep track of the frequency of each letter
    letter_freq = {}
    # Initialize a variable to keep track of the longest balanced substring found so far
    longest_balanced = ""
    # Initialize two pointers, left and right, to keep track of the current substring being considered
    left, right = 0, 0
    # Iterate through the string
    while right < len(s):
        # Update the letter frequency dictionary with the current character
        char = s[right]
        if char.lower() not in letter_freq:
            letter_freq[char.lower()] = 1
        else:
            letter_freq[char.lower()] += 1
        # Move the right pointer to the next character
        right += 1
        # Check if the current substring is balanced
        is_balanced = all(letter_freq.get(c, 0) > 0 and letter_freq.get(c.upper(), 0) > 0 for c in letter_freq)
        if is_balanced:
            # If the current substring is balanced, check if it is longer than the longest found so far
            if len(longest_balanced) < right - left:
                longest_balanced = s[left:right]
        else:
            # If the current substring is not balanced, move the left pointer to the next character
            char = s[left]
            if letter_freq[char.lower()] == 1:
                del letter_freq[char.lower()]
            else:
                letter_freq[char.lower()] -= 1
            left += 1
    return longest_balanced
  
def solution(s):
    """
    Given a string `s` of up to 200 characters, returns the longest balanced substring.
    A balanced substring is defined as a sequence of letters where each letter appears both in lower and upper case.

    Args:
        s (str): The input string.

    Returns:
        str: The longest balanced substring.
    """
    # Initialize a variable to keep track of the longest balanced substring found so far
    longest_balanced = ""
    # Iterate through the string
    for i in range(len(s)):
        # Initialize two sets to keep track of the lowercase and uppercase letters seen so far
        lowercase_seen = set()
        uppercase_seen = set()
        for j in range(i, len(s)):
            if s[j].islower():
                lowercase_seen.add(s[j])
            elif s[j].isupper():
                uppercase_seen.add(s[j].lower())
            # Check if the current substring is balanced
            if lowercase_seen == uppercase_seen:
                # If the current substring is longer than the longest found so far, update it
                if len(longest_balanced) < len(s[i:j+1]):
                    longest_balanced = s[i:j+1]
    return len(longest_balanced) if len(longest_balanced)>0 else -1

print(solution('azABaabza'))
print(solution('TacoCat'))
print(solution('TestTES'))
print(solution('aaabcABccccc'))


import paho.mqtt.client as mqtt
mqtt.MQTT_ERR_SUCCESS
from time import sleep

DEVICE_ID = '12345'

_CONFIG_RECEIVER = {
  'MQTT_HOST': 'HOSTNAME',
  'MQTT_PORT': 1883,
  'MQTT_USER': 'USERNAME',
  'MQTT_PASS': 'PASSWORD',
  'MQTT_PATH': 'PATH',
  'MQTT_TOPIC': 'config1',
  'MQTT_QOS': 1
  }

def on_connect(client, userdata, flags, rc):
  print('Connected with result code {}'.format(rc))
  path  = _CONFIG_RECEIVER['MQTT_PATH']
  topic = _CONFIG_RECEIVER['MQTT_TOPIC']
  client.subscribe(topic='{}/{}'.format(path, topic), qos=_CONFIG_RECEIVER['MQTT_QOS'])
  return

def on_disconnect(client, userdata, rc):
  # if rc != 0:
  print('Disconnected with result code {}'.format(rc))
  return

def on_message(client, userdata, message):
  try:
    import base64
    import io
    import numpy as np
    import json
    from PIL import Image
    payload = json.loads(message.payload)
    _id = payload['ID']
    img_base64 = payload['IMG']
    img = base64.b64decode(img_base64)
    img = Image.open(io.BytesIO(img))
    
    lst_received.append(_id)
    
    print('Received message _id {}/{} on topic {} with QoS {}. Total messages: {}'.format(
      _id,
      str(np.array(img).shape), 
      message.topic, 
      str(message.qos),
      len(lst_received)
      ))
  except Exception as e:
    print(str(e))
  return

def on_publish(client, userdata, mid):
  print('Publish {}, {}, {}'.format(client, userdata, mid))
  return
  
def on_log(mqttc, userdata, level, string):
  print(string)
  return
  

if __name__ == '__main__':
  lst_received = []
  
  mqttc = mqtt.Client(
    client_id=DEVICE_ID,
    clean_session=False
    )
  
  mqttc.username_pw_set(
    username=_CONFIG_RECEIVER['MQTT_USER'],
    password=_CONFIG_RECEIVER['MQTT_PASS']
    )
  
  mqttc.on_connect = on_connect
  mqttc.on_disconnect = on_disconnect
  mqttc.on_message = on_message
  mqttc.on_publish = on_publish
  mqttc.on_log = on_log
  
  mqttc.connect(
    host=_CONFIG_RECEIVER['MQTT_HOST'],
    port=_CONFIG_RECEIVER['MQTT_PORT']
    )
  
  # mqttc.loop_forever()
  
  topic_name = '{}/{}'.format(_CONFIG_RECEIVER['MQTT_PATH'], _CONFIG_RECEIVER['MQTT_TOPIC'])
  # while True:
  #   mqttc.loop(timeout=0.01)
  #   sleep(0.3)
  
  # mqttc.loop_forever()
  # mqttc.loop_start()
  
  
  while True:
    mqttc.loop_read()
    # mqttc.loop(timeout=0.01)
    # mqttc.loop_misc()
    sleep(0.1)
      



import json

try:
  from gql import gql, Client
  from gql.transport.requests import RequestsHTTPTransport
  GQL_INSTALL = True
except:
  GQL_INSTALL = False

def gql_to_config(dct_gql, STREAMS_KEY=None):
  if STREAMS_KEY is None:
    STREAMS_KEY = 'pluginStreamConfigByFields'

  if STREAMS_KEY in dct_gql:
    lst_streams = dct_gql[STREAMS_KEY]
  else:
    lst_streams = dct_gql
  assert isinstance(lst_streams, list)
  return lst_streams

if __name__ == '__main__':
  if not GQL_INSTALL:
    raise ValueError("Please run `pip install gql`")
    
  SERVER_ADDR = "https://graphql-server-jx-staging.jenkinsx.globalintelligence.ro/"
  #SERVER_ADDR = "http://graphql.gts-ws-test.globalintelligence.ro"
  QUERY = """
    query PluginStreamConfigByFields($boxId: String) {
        pluginStreamConfigByFields(boxId: $boxId) {
          NAME          
          DESCRIPTION          
          LIVE_FEED          
          RECONNECTABLE          
          TYPE          
          URL          
          STREAM_CONFIG_METADATA {          
            DATA_HELPER_NAME
            DATA_HELPER_PARAMS
            TRANSCODER_H
            TRANSCODER_W          
            }      
          CAP_RESOLUTION
          PLUGINS {        
        SIGNATURE        
        INSTANCES {
          PLUGIN_INSTANCE_PARAMETER_LIST {
            NAME        
            VALUE        
            TYPEVALUE
            }
          }
        }
        }
      }
    """
  VARIABLE_VALUES = {
    "boxId": None#"gts_ws_test"
    }
    
  STREAMS_KEY = 'pluginStreamConfigByFields'
    
  
  http_transp=RequestsHTTPTransport(
      url=SERVER_ADDR,
      use_json=True,
      headers={
          "Content-type": "application/json",
      },
      verify=False,
      retries=3,
  )
  
  client = Client(
      transport=http_transp,
      fetch_schema_from_transport=True,
  )
  
  query = gql(QUERY)
  
  resp = client.execute(query, variable_values=VARIABLE_VALUES)
  
  streams = resp[STREAMS_KEY]
  print('**********************************\n'*3)
  s_streams = json.dumps(streams, indent=4)
  print(s_streams)
  
  converted_streams = gql_to_config(streams)

  print('**********************************\n'*3)
  s_streams = json.dumps(converted_streams, indent=4)
  print(s_streams)
      
  


from multiprocessing import Process, Pipe
import time

def reader_proc(pipe):
    ## Read from the pipe; this will be spawned as a separate Process
    p_output, p_input = pipe
    p_input.close()    # We are only reading
    while True:
        msg = p_output.recv()    # Read from the output pipe and do nothing
        if msg=='DONE':
            break

def writer(count, p_input):
    for ii in range(0, count):
        p_input.send(ii)             # Write 'count' numbers into the input pipe
    p_input.send('DONE')

if __name__=='__main__':
    for count in [10**4, 10**5, 10**6]:
        # Pipes are unidirectional with two endpoints:  p_input ------> p_output
        p_output, p_input = Pipe()  # writer() writes to p_input from _this_ process
        reader_p = Process(target=reader_proc, args=((p_output, p_input),))
        reader_p.daemon = True
        reader_p.start()     # Launch the reader process

        p_output.close()       # We no longer need this part of the Pipe()
        _start = time.time()
        writer(count, p_input) # Send a lot of stuff to reader_proc()
        p_input.close()
        reader_p.join()
        print("Sending {0} numbers to Pipe() took {1} seconds".format(count,
            (time.time() - _start)))


# TODO Bleo: WIP
from typing import List, Tuple, Union, Any

import numpy as np
import torch as th
from naeural_core.local_libraries.nn.th.training.data.base import BaseDataLoaderFactory
from naeural_core.local_libraries.nn.th.training_utils import read_image
from naeural_core.local_libraries.nn.th.image_dataset_stage_preprocesser import PreprocessResizeWithPad, PreprocessMinMaxNorm


class AutoencoderDataLoaderFactory(BaseDataLoaderFactory):

  def __init__(self, image_size, **kwargs):
    self.image_height, self.image_width = image_size
    super(AutoencoderDataLoaderFactory, self).__init__(**kwargs)
    return

  def startup(self):
    super().startup()
    return

  def _lst_on_load_preprocess(self) -> List[Tuple[Any, dict]]:
    return [
      (PreprocessResizeWithPad, dict(h=self.image_height, w=self.image_width, normalize=False))
    ]

  def _lst_right_before_forward_preprocess(self) -> List[Tuple[Any, dict]]:
    return [
      (PreprocessMinMaxNorm, dict(min_val=0, max_val=255, target_min_val=0, target_max_val=1))
    ]

  def _get_not_loaded_observations_and_labels(self) -> Tuple[Union[List, np.ndarray], Union[List, np.ndarray]]:
    all_paths = self.dataset_info['paths']
    image_paths = [path for path in all_paths if path.endswith(('.png', '.jpg', '.jpeg'))]
    observations = image_paths
    labels = image_paths
    return observations, labels

  def _load_x_and_y(self, observations, labels, idx) -> Tuple[Union[th.tensor, np.ndarray], Union[th.tensor, np.ndarray]]:
    x = th.tensor(read_image(observations[idx]))
    y = th.tensor(read_image(labels[idx]))
    return x,y

  def _preprocess(self, x, y, transform):
    x = transform(x)
    y = transform(y)
    return x,y

if __name__ == '__main__':

  from naeural_core import SBLogger

  log = SBLogger()
  train_data_factory = AutoencoderDataLoaderFactory(
    log=log,
    num_workers=0,
    batch_size=32,
    path_to_dataset='./_local_cache/_data/MNIST/dev',
    data_subset_name='dev',
    image_height=None,
    image_width=None,
    device='cpu',
    files_extensions=None,
    prc_noise=0.3
  )
  train_data_factory.create()

  for step, data in enumerate(train_data_factory.data_loader):
    break







import pyodbc
import pandas as pd

from naeural_core.local_libraries.db_conn.base import BaseConnector

class ODBCConnector(BaseConnector):
  def __init__(self, **kwargs):
    self._cnxn = None
    self._default_sql_query = "SELECT * FROM {};"
    super(ODBCConnector, self).__init__(**kwargs)
    return

  def _connect(self, **kwargs):
    str_conn = ''
    for k,v in kwargs.items():
      str_conn += '{}={};'.format(k,v)
    self._cnxn = pyodbc.connect(str_conn)
    return

  def check_connection(self):
    try:
      cursor = self._cnxn.cursor()
      cursor.execute("SELECT 1")
    except pyodbc.Error as e:
      return False
    return True

  def _create_reader(self, **kwargs):
    table_data = kwargs.get('table_data', '')
    chunksize = kwargs.get('chunksize', None)
    sql_query = kwargs.get('sql_query', '')
    assert len(table_data) > 0 or len(sql_query) >0, "Must either have table or SQL query"
    if len(sql_query) == 0:      
      sql_query = self._default_sql_query.format(table_data)

    reader = pd.read_sql(
      sql=sql_query,
      con=self._cnxn,
      chunksize=chunksize
     )
    return reader




if __name__ == '__main__':

  from naeural_core import Logger

  log = Logger(
    lib_name='DB', base_folder='dropbox', app_folder='_lens_data/_product_dynamics',
    TF_KERAS=False
  )

  config = {
    'CONNECT_PARAMS' : {
      'DRIVER' : '{ODBC Driver 17 for SQL Server}',
      'SERVER' : 'cloudifiersql1.database.windows.net',
      'PORT' : 1433,
      'DATABASE' : 'operational',
      'Uid' : 'damian',
      'Pwd' : 'MLteam2021!',
      'Encrypt' : 'yes',
      'TrustServerCertificate' : 'no',
      'Connection Timeout': 30,
    },

    'QUERY_PARAMS' : {
      'default' : {
        'table_data' : 'Invoices',
        'sql_query' : "", ### custom sql query on 'TABLE_DATA' (groupby etc etc); if empty it uses a default sql query
        'chunksize' : 200, ### if removed, then the generator conn.data_chunk_generator() will have only one step
      },

      'default2' : {
        'table_data' : 'Invoices',
        "sql_query" : "",
        'chunksize' : 200,
      }
    }
  }

  conn = ODBCConnector(log=log, config=config)
  conn.connect(nr_retries=5)
  dct_data = conn.get_all_data()